"""
Granular analysis methods (eg. D<sup>2</sup><sub>min</sub>), network analysis methods (eg. adjacency matrices), and inverse force solving.
"""

from .BrightnessGSquared import *
from .NetworkAnalysis import *
from .ForceBalance import *
from .ForceSolve import *
from .D2Min import *
from .Derivative import *
from .LinearRegions import *



"""
Defines the data type of each argument for the circle tracking methods.
"""

circleTrackArgDTypes = {"radiusTolerance": int,
                     "offscreenParticles": bool,
                     "outlineOnly": bool,
                     "outlineThickness": float,
                     "negativeHalo": bool,
                     "haloThickness": float,
                     "negativeInside": bool,
                     "peakDownsample": int,
                     "minPeakPrevalence": float,
                     "intensitySoftmax": float,
                     "intensitySoftmin": float,
                     "invert": bool,
                     "allowOverlap": bool,
                     "fitPeaks": bool,
                     "debug": bool}




"""

Methods for visualizing the results of other submodules, like displaying particle
positions, forces, or contact angles.

Listed below are some important tips and conventions that are
used (both within this submodule and externally) that will be
helpful to keep in mind as you are working with this library.

Visualization Tips
------------------

### Coordinate Order

Nearly every part of this library will return coordinates as `[y,x]`,
which is the reverse of the typical order `[x,y]`. This is done because
the majority of image processing techniques/analysis methods utilize the
`[y,x]` ordering, and so we stick to it as well. You will also be expected to
input most quantities in this order eg. the center coordinates for a circular
mask created with `pepe.preprocess.circularMask()`.

Matplotlib, on the other hand, normally assumes `[x,y]` ordering, and so it
may be necessary to flip the order upon graphing any positions. This is very
easy to do with array slicing:

```
a = np.array([[1, 2], [3, 4], [5, 6]])

# Normal order
print(a) # Output: [[1, 2], [3, 4], [5, 6]]

# Flipped order
print(a[:,::-1]) # Output: [[2, 1], [4, 3], [6, 5]]
```

### Flipped Y-Axis

Another convention of images is to invert the y-axis when displaying an
image, such that `[0,0]` is in the top-left corner of the plot. This is
the case whenever you display something with `plt.imshow()`, but will not
be implemented if you simply plot circles with `plt.Circle()` (or similar).

As a result, it may be necessary to flip the axis manually with:

```
ax.invert_yaxis()
```

If this is not done, the positions of contact points may look inverted, which
can be especially hard to notice if you system has vertical symmetry (since
it won't be obvious by just the positions of the particles).
"""

from .Colors import *
from .Circles import *
from .Forces import *
from .Rotation import *



"""
String-to-list parsing used in reading settings files.
"""
import numpy as np
import ast

def parseList(string, dtype=object):
    """
    Takes in a string representation of a list
    and builds a proper list.
    """
    # Most of the heavy lifting is done by ast 
    builtList = ast.literal_eval(string)

    # Now we can use numpy to fix the type
    builtArr = np.array(builtList, dtype=dtype)

    # Now convert back to a list
    return list(builtList)



"""
Methods to visualize rotational motion of a particle.
"""
import numpy as np
import matplotlib.pyplot as plt

from PIL import Image

from pepe.preprocess import checkImageType
from pepe.visualize import genColors, visCircles

def visRotation(imageFilePaths, centerArr, radiusArr, angleArr, outputFolderPath='./', offset=(0,0), barWidth=2.5, outputFileName='particle_orientation.gif'):
    """
    Create a gif of the tracked (position and orientation) particles
    overlaid on the original images.
    """

    # Where we will store each drawn image/figure, to be made into
    # a gif afterwards
    imageArr = [None]*len(imageFilePaths)

    colorArr = genColors(len(centerArr))

    for i in range(len(imageFilePaths)):
        
        fig, ax = plt.subplots()

        image = checkImageType(imageFilePaths[i])
        ax.imshow(image)

        # Draw the circles
        visCircles([centerArr[p][i] + np.array(offset) for p in range(len(centerArr))], [radiusArr[p][i] for p in range(len(radiusArr))], ax=ax, colors=colorArr)

        # Now we want to draw a bar in the center of each circle to
        # visualize the rotation
        for p in range(len(centerArr)):
            # Left end of the line
            point1 = centerArr[p][i] + np.array(offset) - np.array([np.sin(angleArr[p][i]), np.cos(angleArr[p][i])]) * radiusArr[p][i] / 2
            # Right end of the line
            point2 = centerArr[p][i] + np.array(offset) + np.array([np.sin(angleArr[p][i]), np.cos(angleArr[p][i])]) * radiusArr[p][i] / 2

            # Now actually draw the line
            plt.plot([point1[1], point2[1]], [point1[0], point2[0]], linewidth=barWidth, c=colorArr[p])


        canvas = plt.get_current_fig_manager().canvas
        canvas.draw()
        imageArr[i] = Image.frombytes('RGB', canvas.get_width_height(), 
                                    canvas.tostring_rgb())

        plt.close(fig)


    imageArr[0].save(outputFolderPath + outputFileName, save_all=True,
                                append_images=imageArr[1:], duration=25, optimize=True, loop=True)

    return




"""
These methods aren't really 'topological', but I think this is
still the most apt place to put this code.
"""

import numpy as np
from scipy.spatial import KDTree

def spatialClusterLabels(points, l=.001):
    """
    Partition a set of points in clusters, and return the
    cluster label for each point.

    Method is very simplistic: the neighbors (points within
    a threshold distance) of a point are added to the same
    cluster, and the process is repeated until no points
    remain. This means that the size of clusters can be
    larger than this distance threshold (eg. think of a
    chain of points).

    Parameters
    ----------
    points : numpy.ndarray[N,d]
        Array of N points in d-dimensions

    l : float
        Distance threshold to consider two points as being in the
        same cluster. Given as fraction of the total system size.
        
    Returns
    -------
    labels : numpy.ndarray[N]
        Array of N labels (integers) denoting cluster
        assignment. Total number of clusters will be `max(labels)+1`.
    """
    d = np.shape(points)[-1]
    
    systemLengthScale = np.sqrt(np.sum([(np.max(points[:,i]) - np.min(points[:,i]))**2 for i in range(d)]))
    threshold = l*systemLengthScale
    
    # Generate a kd-tree
    kdTree = KDTree(points)

    pointsList = points.tolist()
    randomOrder = np.arange(len(pointsList))
    np.random.shuffle(randomOrder)
    
    labels = np.zeros(len(pointsList)) - 1
    labelsToMerge = []
    
    i = 0
    
    while len(np.where(labels == -1)[0]) > 0:
        # Grab a point, if it doesn't have a group already, create a new one
        if labels[randomOrder[i]] == -1:
            labels[randomOrder[i]] = np.max(labels)+1
        p = pointsList[randomOrder[i]]

        # Find its neighbors
        neighbors = kdTree.query_ball_point(p, r=threshold)

        # neighorLabels = [labels[n] for n in neighors]
        # uniqueNeighborLabels = np.unique(neighborLabels)
        # if len(uniqueNeighborLabels) == 1:
        #     # If all are unlabeled
        #     if uniqueNeighborLabels[0] == -1:
        #         # Create a new group
        #         labels[randomOrder[i]] = np.max(labels)+1

        #     # All are already in another group
        #     else:
        #         # Assign this point to that group as well
        #         labels[randomOrder[i]] = uniqueNeighborLabels[0]
        
        # Assign all of them to be the same group
        for index in neighbors:
            if index == randomOrder[i]:
                continue
                
            # If they already have a label, we will eventually want to merge the labels
            if labels[index] >= 0 and labels[index] != labels[randomOrder[i]]:
                # See if this is already in a merge group
                inGroup = False
                for j in range(len(labelsToMerge)):
                    # If it is already going to be merged, add the
                    # new label to this group
                    if labels[index] in labelsToMerge[j]:
                        inGroup = True
                        # If the new label is not in the merge group, add it
                        if not labels[randomOrder[i]] in labelsToMerge[j]:
                            labelsToMerge[j].append(labels[randomOrder[i]])
                            
                        break
                # If we aren't in a merge group, we should create
                # a new one
                if not inGroup:
                    labelsToMerge.append([labels[index], labels[randomOrder[i]]])
                    
            elif labels[index] == -1:
                labels[index] = labels[randomOrder[i]]
            
        i += 1

    mergedLabels = np.zeros_like(labels) - 1
    
    # Now we have to sort out merged clusters
    labelSets = []
    for i in range(int(np.max(labels))+1):
        # Generate a list of all of the labels that are identified together
        # including this value of i
        allLabels = [i]
        for j in range(len(labelsToMerge)):
            if i in labelsToMerge[j]:
                allLabels += labelsToMerge[j]

        # sort so that way we can do a unique check to remove
        # duplicates
        allLabels = sorted(np.unique(allLabels))
        # Python may automatically collapse the list if it is of length 1,
        # which we don't want.
        labelSets.append([allLabels] if len(allLabels) == 1 else allLabels)

    # Remove duplicate sets, creating disjoint sets of labels
    uniqueSets = np.unique(np.array(labelSets, list))

    # If there are no overlaps, then this above operation will
    # flatten the array to be one dimensional. We can't do anything
    # about that, except to check if it is the case. The good news is
    # that if this is the case, then we already have disjoint sets
    # and we can just immediately return our label array
    if len(np.shape(uniqueSets)) == 1:
        return labels

    # Relabel things with the new disjoint sets
    mergedLabels = np.zeros_like(labels)

    # TODO: Probably a much better way to do this
    for i in range(len(pointsList)):
        for j in range(len(uniqueSets)):
            if labels[i] in uniqueSets[j]:
                mergedLabels[i] = j
                break

    return mergedLabels

def spatialClusterCenters(points, l=.001, return_weights=False):
    """
    Partition a set of points in clusters, and compute the
    center of each cluster.

    Generates clusters using `pepe.topology.spatialClusterLabels()`.

    Parameters
    ----------
    points : numpy.ndarray[N,d]
        Array of N points in d-dimensions

    l : float
        Distance threshold to consider two points as being in the
        same cluster. Given as fraction of the total system size.

    return_weights : bool
        Whether to return the weight -- defined as the fraction of
        points included in that cluster -- alongside the centers.
        
    Returns
    -------
    centers : numpy.ndarray[N, d]
        Array of N points in d-dimensions representing the detected
        clusters in the system.

    weights : numpy.ndarray[N]
        Array of weights -- defined as fraction of all points included
        in each cluster -- for each cluster. Only returned if
        `return_weights=True`.

    """
    labels = spatialClusterLabels(points, l=l)
    numLabels = int(np.max(labels))+1
    
    # Compute the center of each cluster
    weights = np.zeros(numLabels)
    centers = np.zeros((numLabels, np.shape(points)[-1]))
    
    for i in range(numLabels):
        indices = np.where(labels == i)
        centers[i] = np.mean(np.array(points)[indices], axis=0)
        weights[i] = len(indices[0])/len(labels)

    order = np.argsort(weights)[::-1]
    centers = centers[order]
    weights = weights[order]
    
    if return_weights:
        return centers, weights
        
    return centers



"""
Miscellaneous functions, like coordinate transformations, function parsing, and proximity sorting.
"""
from .Outer import *
from .Sort import *
from .RectArr import *
from .Kwargs import *
from .FittingFunctions import *
from .Parse import *
from .Coordinates import *



"""
Methods to visualize the positions of particles.
"""
import numpy as np
import matplotlib.pyplot as plt

from pepe.visualize import genColors

def visCircles(centers, radii, ax=None, colors=None, annotations=None, sameColors=False, setBounds=False, linewidth=1):
    """
    Draw circles on an axis.

    Parameters
    ----------

    centers : np.ndarray[N,2] or list[N,2]
        List of particle centers [y,x] (in pixels).

    radii : np.ndarray[N] or list[N]
        List of particle radii (in pixels).
    """
    # Cut out nan values
    goodIndices = np.array([0 if np.isnan(r) else 1 for r in radii])
    npRadii = np.array([r for r in radii if not np.isnan(r)])
    npCenters = np.array([c for c in centers if not np.isnan(c[0])])

    if ax is None:
        fig, ax = plt.subplots()

    if colors is None:
        if sameColors:
            singleColor = genColors(1)[0]
            circleColorsList = [singleColor for _ in range(len(npCenters))]
        else:
            circleColorsList = genColors(len(npCenters))

    elif not type(colors) is list:
        # If a single color is given, we want to repeat that
        circleColorsList = [colors for _ in range(len(npCenters))]

    else:
        circleColorsList = colors

    for i in range(len(npCenters)):
        c = plt.Circle(npCenters[i,::-1], npRadii[i], color=circleColorsList[i], fill=False, linewidth=linewidth)
        ax.add_artist(c)

    if annotations is not None:
        if len(annotations) != len(centers):
            print('Warning: Invalid number of annotations provided! Ignoring...')
            pass
        else:
            for i in range(len(npCenters)):
                # The annotation array might be mismatched if we had any nan circles
                # so we have to account for that
                ax.text(npCenters[i,1], npCenters[i,0], annotations[i + int(np.sum(1 - goodIndices[:i]))], ha='center', va='center', color=circleColorsList[i])

    if setBounds:
        ax.set_xlim([np.min(npCenters[:,1])-1.5*np.max(npRadii), np.max(npCenters[:,1])+1.5*np.max(npRadii)])
        ax.set_ylim([np.min(npCenters[:,0])-1.5*np.max(npRadii), np.max(npCenters[:,0])+1.5*np.max(npRadii)])
        ax.set_aspect('equal')

    return ax



"""
Tracking particles and forces across timesteps to turn triangular arrays into rectangular ones.
"""
import numpy as np
import matplotlib.pyplot as plt

from pepe.utils import preserveOrderArgsort

def rectangularizeForceArrays(forceArr, alphaArr, betaArr, centerArr, radiusArr, maxBetaDisplacement=.3, forcePadValue=np.nan, continuousBetas=False):
    """
    Take the ragged arrays representing particles and forces and convert them to
    particle-wise rectangular arrays.

    The force input arrays should be 3 dimensional, with indices as follows:

        `forceArr[timestep index]
                [particle index]
                [force index]`

    The center array should be 3 dimensional as well, with indices as follows:

        `centerArr[timestep index]
                 [particle index]
                 [coordination (x or y) index]`

    The radius array should be 2 dimensional:

        `radiusArr[timestep index]
                 [particle index]`

    The final returned result will be a list of rectangular arrays of the form:

        `rectForceArr[particle index]
                    [force index]
                    [timestep index]`

    There are several advantages to restructing like this:

        1.  The evolution of a single force can be plotted with something
            like `plt.plot(rectForceArr[0][0])`. To do this with the input arrays
            would require quite a bit more work since they are lists, and cannot
            be sliced.
        
        2.  While the top-layer index corresponds to a python list, all of the particle-
            wise arrays are numpy objects, which allows them to be sliced/saved easily.

        3.  Due to potentially having different numbers of particles per frame and different
            numbers of forces acting on each particle, the original arrays and very ragged.
            This makes any sane indexing very hard to do, which is fixed by reshaping the
            arrays. If a force only exists for a subset of the time steps, the rectangular
            result will still identify it as a unique force in all timesteps, just with a
            magnitude of 0 (or `np.nan`; see `forcePadValue`).


    Parameters
    ----------

    forceArr : list[T,P,F]
        Force magnitudes for T timesteps, with P(T) particles per timestep, and F(P(T)) forces
        per particle.

    alphaArr : list[T,P,F]
        Alpha angles for T timesteps, with P(T) particles per timestep, and F(P(T)) forces
        per particle.

    betaArr : list[T,P,F]
        Beta angles for T timesteps, with P(T) particles per timestep, and F(P(T)) forces
        per particle.

    centerArr : list[T,P,2]
        List of particle centers (y,x) for T timesteps, with P(T) particles per timestep.

    radiusArr : list[T,P]
        List of particle radii for T timesteps, with P(T) particles per timestep.

    angleArr : list[T,P]
        List of particle rotations for T timesteps, with P(T) particles per timestep.

    maxBetaDisplacement : float
        The maximum difference of beta angle between timesteps for which a force will
        still be identified with the previous step. The difference is calculated between
        the current timestep and whatever the last timestep for which that force existed.

    forcePadValue : float
        The value that is recorded for a force that isn't current active for a given frame.
        Common choices would be `np.nan` -- such that the force won't appear when it is
        plotted -- or `0` -- which is a little more physically-inspired.

    continuousBetas : bool
        When a force is not active for a number of timesteps, should the beta value (position)
        of that force be set to `np.nan` (`False`) or should the previous position of the
        force be carried over (`True`)? This is the difference between the force not existing
        for these timesteps, versus the force existing but having a magnitude of 0.

    Returns
    -------

    rectForceArr : list[P](np.ndarray[F,T])
        A list of arrays representing the force magnitude for each force on each particle.
        See description for more info.

    rectAlphaArr : list[P](np.ndarray[F,T])
        A list of arrays representing the alpha angle for each force on each particle.
        See description for more info.

    rectBetaArr : list[P](np.ndarray[F,T])
        A list of arrays representing the beta angle for force on each particle.
        See description for more info.

    rectCenterArr : np.ndarray[P,T,2]
        Particle centers for each timestep. Elements take on a value of `[np.nan, np.nan]`
        if the particle does not exist for a given timestep.

    rectRadiusArr : np.ndarray[P,T]
        Particle radii for each timestep. Elements take on a value of `np.nan` if the particle
        does not exist for a given timestep.

    rectAngleArr : np.ndarray[P,T]
        Particle angles for each timestep. Elements take on a value of `np.nan` if the particle
        does not exist for a given timestep.
    """
    rectForceArr = []
    rectBetaArr = []
    rectAlphaArr = []

    # Scalar
    # TODO Make this capable of following same process as for forces, where
    # identities are maintained even if multiple particles appear/disappear
    # To determine the maximum number of particles, we have to go through each frame
    # and give unique identities to each one. We can't just take the max length of the
    # list, since one frame could have 3 particles, and another 2 particles, but that
    # could be a total of 5 unique particles (or it could be 3 unique particles).
    
    numTimesteps = len(forceArr)

    # We'll build this list as we go, which will be triangular (ie. a 2d list in which
    # each sublist could have a different number of elements). That being said, by keeping
    # placeholders for particles that disappear, we will ensure that this is upper triangular,
    # meaning it will only every grow in size.

    # Initialize the first step
    triSortedParticleOrderArr = [[i for i in range(len(centerArr[0]))]]
    triSortedCenterArr = [[c for c in centerArr[0]]]

    for j in range(1, numTimesteps):

        pOrder = preserveOrderArgsort(triSortedCenterArr[-1], centerArr[j], padMissingValues=True, fillNanSpots=True, maxDistance=np.mean(radiusArr[j]))
        # If a particle disappears, we carry through the old center, so it can be identified again
        # in the future
        centers = [centerArr[j][pOrder[k]] if pOrder[k] is not None else triSortedCenterArr[-1][k] for k in range(len(pOrder))]

        triSortedCenterArr.append(centers)
        triSortedParticleOrderArr.append(pOrder)

    #print(triSortedParticleOrderArr)

    maxNumParticles = len(triSortedCenterArr[-1])

    # First, make the centers array look nice, which we then use to identify
    # particles
    rectCenterArr = np.zeros((maxNumParticles, numTimesteps, 2))
    rectRadiusArr = np.zeros((maxNumParticles, numTimesteps))
    particleOrder = np.zeros((numTimesteps, maxNumParticles), dtype=np.int16)

    for j in range(numTimesteps):
        particleOrder[j] = [pI if pI is not None else -1 for pI in triSortedParticleOrderArr[j]] + [-1]*(maxNumParticles - len(triSortedParticleOrderArr[j]))
        # Now just copy the triangular data into a rectangular array, padding at the end as necessary
        # (and replacing invalid data with np.nan as necessary).
        rectCenterArr[:,j] = [triSortedCenterArr[j][k] if triSortedParticleOrderArr[j][k] is not None else [np.nan, np.nan] for k in range(len(triSortedParticleOrderArr[j]))] + [[np.nan, np.nan] for k in range(maxNumParticles - len(triSortedParticleOrderArr[j]))]
        rectRadiusArr[:,j] = [radiusArr[j][triSortedParticleOrderArr[j][k]] if triSortedParticleOrderArr[j][k] is not None else np.nan for k in range(len(triSortedParticleOrderArr[j]))] + [np.nan for k in range(maxNumParticles - len(triSortedParticleOrderArr[j]))]

    # We have to initialize the first element so that we can then use the
    # preserveOrderSort function to make sure the identities stay
    # consistent
#    rectCenterArr[:,0] = list(centerArr[0]) + [[np.nan, np.nan]]*(maxNumParticles - len(centerArr[0]))
#    rectRadiusArr[:,0] = list(radiusArr[0]) + [np.nan]*(maxNumParticles - len(centerArr[0]))
#
#    particleOrder = np.zeros((numTimesteps, maxNumParticles), dtype=np.int16)
#    particleOrder[0] = np.arange(len(particleOrder[0]))
#    particleExists = np.zeros((numTimesteps, maxNumParticles), dtype=np.int16)
#
#    for i in range(1, numTimesteps):
#        currOrder = preserveOrderArgsort(rectCenterArr[:,i-1], centerArr[i], padMissingValues=True, fillNanSpots=True, maxDistance=np.mean(rectRadiusArr[:,i-1]))
#        # Convert all None to -1 (since None can't be turned into an integer)
#        particleOrder[i] = [ci if ci is not None else -1 for ci in currOrder]
#        rectCenterArr[:,i] = [centerArr[i][particleOrder[i,j]] if particleOrder[i,j] >= 0 else [np.nan, np.nan] for j in range(len(particleOrder[i]))]
#        rectRadiusArr[:,i] = [radiusArr[i][particleOrder[i,j]] if particleOrder[i,j] >= 0 else np.nan for j in range(len(particleOrder[i]))]


    # We now have linked the particles from frame to frame, and can
    # rectangularize the other quantities on a particle-by-particle basis
    for i in range(maxNumParticles):
        # Unfortunately, we cannot determine the maximum number of forces just by doing a
        # call to np.max(len(...)) or something, because eg. two frames could have 3 and 4
        # forces in each, but those could represent up to 7 unique forces total

        # This will be triangular array representing the indices of each force in the final
        # rectangular array
        # We have to make sure this particle exists at the first timestep as well
        #print(particleOrder[0])
        #print(betaArr[0])
        if particleOrder[0,i] >= 0 and particleOrder[0,i] < len(betaArr[0]):
            triSortedBetaArr = [betaArr[0][particleOrder[0,i]]]
            triSortedForceOrderArr = [list(np.arange(len(triSortedBetaArr[0])))]
        else:
            # Otherwise, we just use an empty list
            triSortedBetaArr = [[]]
            triSortedForceOrderArr = [[]]


        for j in range(1, numTimesteps):
            particleIndex = particleOrder[j,i]
   
            # If less than 0, that means this particle is not actually present in this
            # frame, so we just continue. Note that we do have to carry forward the
            # beta values from the previous frame, since otherwise we will run into
            # issues later on.
            if particleIndex < 0:
                triSortedBetaArr.append(triSortedBetaArr[-1])
                triSortedForceOrderArr.append([None for i in range(len(triSortedBetaArr[-1]))])
                continue

            # Correlate the forces between this frame and the last
            #print(triSortedBetaArr[-1])
            order = preserveOrderArgsort(triSortedBetaArr[-1], betaArr[j][particleIndex], padMissingValues=True, maxDistance=maxBetaDisplacement, periodic=True)
            # If we "lose" a force, carry the old beta value through to the new array, so
            # that we can identify it if it ever pops up again.
            betaValues = [betaArr[j][particleIndex][order[k]] if order[k] is not None else triSortedBetaArr[-1][k] for k in range(len(order))]

            #print(f'{betaValues}     {order}     {betaArr[j][particleIndex]}')

            triSortedBetaArr.append(betaValues)
            #triSortedForceOrderArr.append([i for i in order if i is not None])
            triSortedForceOrderArr.append(order)

        #plt.plot([triSortedBetaArr[j][0] for j in range(len(triSortedBetaArr))])
        #plt.show()

        # Since the triangular array is only allowed to grow, the length of the last
        # entry will be the maximum number of *unique* forces.
        #plt.plot([len(ele) for ele in triSortedBetaArr]) # Check that it only grows
        #plt.show()
        maxNumForces = len(triSortedBetaArr[-1])

        if maxNumForces == 0:
            continue

        # Now we can finally create some nice rectangular arrays
        singleParticleForceArr = np.zeros((maxNumForces, numTimesteps))
        singleParticleBetaArr = np.zeros((maxNumForces, numTimesteps))
        singleParticleAlphaArr = np.zeros((maxNumForces, numTimesteps))

        for j in range(numTimesteps):
            particleIndex = particleOrder[j,i]
            
            # If particle doesn't exist for this frame, fill with nan values and move
            # on
            if particleIndex < 0:
                singleParticleBetaArr[:,j] = np.repeat(np.nan, maxNumForces)
                singleParticleForceArr[:,j] = np.repeat(forcePadValue, maxNumForces)
                singleParticleAlphaArr[:,j] = np.repeat(np.nan, maxNumForces)
                continue

            # We want to remove the carry over beta values for forces that aren't actually there
            # now, and then pad with nan values for forces that haven't appeared yet

            # Not sure why, but it may be helpful to have the previous beta values carried over,
            # so that can be done by passing a kwarg
            if continuousBetas:
                singleParticleBetaArr[:,j] = [triSortedBetaArr[j][k] for k in range(len(triSortedBetaArr[j]))] + [np.nan]*(maxNumForces - len(triSortedBetaArr[j]))
            else:
                singleParticleBetaArr[:,j] = [betaArr[j][particleIndex][triSortedForceOrderArr[j][k]] if triSortedForceOrderArr[j][k] is not None else np.nan for k in range(len(triSortedBetaArr[j]))] + [np.nan]*(maxNumForces - len(triSortedBetaArr[j]))

            # I apologize to anyone who has to read these lines... They do exactly what is stated above,
            # but boy do they look terrible
            singleParticleForceArr[:,j] = [forceArr[j][particleIndex][triSortedForceOrderArr[j][k]] if triSortedForceOrderArr[j][k] is not None else forcePadValue for k in range(len(triSortedBetaArr[j]))] + [forcePadValue]*(maxNumForces - len(triSortedBetaArr[j]))
            singleParticleAlphaArr[:,j] = [alphaArr[j][particleIndex][triSortedForceOrderArr[j][k]] if triSortedForceOrderArr[j][k] is not None else np.nan for k in range(len(triSortedBetaArr[j]))] + [np.nan]*(maxNumForces - len(triSortedBetaArr[j]))

        # Append all of the nice rectangular arrays
        rectForceArr.append(singleParticleForceArr)
        rectAlphaArr.append(singleParticleAlphaArr)
        rectBetaArr.append(singleParticleBetaArr)

    return rectForceArr, rectAlphaArr, rectBetaArr, rectCenterArr, rectRadiusArr




"""
Particle orientation (rotation) tracking via convolution.
"""
import numpy as np

from PIL import Image

import matplotlib.pyplot as plt

def angularConvolution(image, kernelImg, dTheta=.01, angleBounds=(0, 2*np.pi)):
    """
    Perform an angular convolution, calculating the effective similarity between two
    images as one of them is rotated.

    Parameters
    ----------

    image : np.ndarray[H,W] or PIL.Image
       The image for which the rotation should be identified in reference
       to the kernel.

       Should be a single channel image; if a multichannel image is passed,
       the convolution will be performed on the image with the channels
       averaged.

    kernelImg : np.ndarray[H,W] or PIL.Image
        The kernel that will be rotated to be matched up with `image`.

       Should be a single channel image; if a multichannel image is passed,
       the convolution will be performed on the image with the channels
       averaged.

    dTheta : float
        The angle stepsize between subsequent rotations, in radians.

        This will only increase the precision up to a certain limit, determined
        by the resolution of the images.

    angleBounds : tuple(float)
        The minimum (first element) and maximum (second element) angle to rotate
        the kernel by during the convolution, in radians.

    Returns
    -------

    angleArr : np.ndarray
        The angles for which the convolution was calculated over, in radians.

    convArr : np.ndarray
        The similarity between the images at each rotation angle.

    """
    # If multiple channels are passed, we grayscale the image
    if len(np.array(image).shape) == 3:
        grayImage = np.mean(image, axis=-1)
    else:
        grayImage = np.array(image)

    if len(np.array(kernelImg).shape) == 3:
        grayKernel = np.mean(kernelImg, axis=-1)
    else:
        grayKernel = np.array(kernelImg)
    
    # Make sure our images are the same size (and make them so if not)
    paddedHeight = max(grayImage.shape[0], grayKernel.shape[0])
    paddedWidth = max(grayImage.shape[1], grayKernel.shape[1])

    paddedImage = np.zeros((paddedHeight, paddedWidth))
    paddedKernel = np.zeros((paddedHeight, paddedWidth))

    imageSizeDiff = np.array([(paddedHeight - grayImage.shape[0])/4, (paddedWidth - grayImage.shape[1])/4], dtype=np.int64)
    kernelSizeDiff = np.array([(paddedHeight - grayKernel.shape[0])/4, (paddedWidth - grayKernel.shape[1])/4], dtype=np.int64)

    # Add in the negatives for the end
    imagePadding = [imageSizeDiff[0], -imageSizeDiff[0], imageSizeDiff[1], -imageSizeDiff[1]]
    kernelPadding = [kernelSizeDiff[0], -kernelSizeDiff[0], kernelSizeDiff[1], -kernelSizeDiff[1]]

    # Change zeros to be None
    imagePadding = [pad if pad > 0 else None for pad in imagePadding]
    kernelPadding = [pad if pad > 0 else None for pad in kernelPadding]

    paddedImage[imagePadding[0]:imagePadding[1], imagePadding[2]:imagePadding[3]] = grayImage
    paddedKernel[kernelPadding[0]:kernelPadding[1], kernelPadding[2]:kernelPadding[3]] = grayKernel


    steps = abs(int((angleBounds[0] - angleBounds[1])/dTheta))
    # Pillow wants angles in degrees, so we convert here (but all of the parameter values
    # should be given in radians)
    angleArr = np.linspace(angleBounds[0]* 180/np.pi, angleBounds[1] * 180/np.pi, steps)
    # We should always make sure that 0 is in our array
    if 0 not in angleArr:
        angleArr = np.append(angleArr, 0)
        angleArr = np.sort(angleArr)
    
    convArr = np.zeros(len(angleArr))
   
    pilKernel = Image.fromarray(paddedKernel)

    for i in range(len(angleArr)):
        convArr[i] = np.sum((paddedImage - np.array(pilKernel.rotate(angleArr[i])))**2)

    # Normalize a bit
    convArr -= np.max(convArr)
    convArr = np.abs(convArr)
        
    return angleArr * np.pi/180, convArr



"""

# <u>P</u>hoto<u>e</u>lastic <u>P</u>ython <u>E</u>nvironment

This is a collection of tools for working with photoelastic particle images, including common analysis methods like particle tracking and community analysis.

The vast majority of these methods are either derivative or directly sourced from previous works (see, e.g., [1] for a nice review of techniques); therefore the value of this library, as I see it, is not in novel implementations of methods, but as a single toolbox containing consistent and compatable implementations of a wide selection of tools.

## Features

- Common analysis techniques (G<sup>2</sup>, D<sup>2</sup><sub>min</sub>, etc.)
- Particle position and orientation tracking
- Masking and other preprocessing tools
- Synthetic photoelastic response generation
- Inverse force solving (as in [2])

## Installation

The library is available on PyPi:
```
pip install pepe-granular
```

It can also be installed from the Github repository:

```
git clone https://github.com/Jfeatherstone/pepe
cd pepe
pip install .
```

Theoretically, this should work for any Python >3, but Python 3.7 and 3.11 have been explicitly tested.

Dependencies:

- [numpy](https://numpy.org/)
- [matplotlib](https://matplotlib.org/)
- [lmfit](https://lmfit.github.io/lmfit-py/index.html)
- [scikit-learn](https://scikit-learn.org/stable/)
- [opencv](https://opencv.org/)
- [Pillow](https://pillow.readthedocs.io/en/stable/)
- [numba](https://numba.pydata.org/)
- [tqdm](https://github.com/tqdm/tqdm)

These can all be installed (alongside their dependencies) via `pip`:
```
git clone https://github.com/Jfeatherstone/pepe
cd pepe
pip install -r requirements.txt
```

Some of the test notebooks may also make use of the Matlab API to compare against Jonathan Kollmer's code, but this is not required to use any functions in the library itself. Installing the Matlab API requires a local installation of Matlab proper; see [here](https://www.mathworks.com/help/matlab/matlab_external/install-the-matlab-engine-for-python.html) for more information.

If you'd rather not install systemwide (say, to run just a simple example), you can also just manually add the package to your path at the top of a Jupyter notebook/python file with (after cloning the repo as above):

```
import sys
sys.path.append('/absolute/path/to/clone/folder/pepe/')
```

Many of the notebooks in the main directory follow this process, since they are used to test the development versions of the packages.

## Usage

See the `pepe.examples` submodule for some common uses of the toolbox. Many pseudo-unit-tests can be found in the repo's [`notebooks`](https://github.com/Jfeatherstone/pepe/tree/master/notebooks) folder, which may also help determine how to use a certain function. Note that these latter examples are more made for my own benefit, and won't be as heavily commented as the proper examples.

## Further Reading and References

[1] Abed Zadeh, A., Barés, J., Brzinski, T. A., Daniels, K. E., Dijksman, J., Docquier, N., Everitt, H. O., Kollmer, J. E., Lantsoght, O., Wang, D., Workamp, M., Zhao, Y., & Zheng, H. (2019). Enlightening force chains: A review of photoelasticimetry in granular matter. Granular Matter, 21(4), 83. https://doi.org/10.1007/s10035-019-0942-2

[2] Daniels, K. E., Kollmer, J. E., & Puckett, J. G. (2017). Photoelastic force measurements in granular materials. Review of Scientific Instruments, 88(5), 051808. https://doi.org/10.1063/1.4983049

[2a] Jonathan Kollmer's implementation in Matlab: https://github.com/jekollmer/PEGS

[2b] Olivier Lantsoght's implementation in Python: https://git.immc.ucl.ac.be/olantsoght/pegs_py

[3] Photoelastic methods wiki. https://git-xen.lmgc.univ-montp2.fr/PhotoElasticity/Main/-/wikis/home
"""

__version__ = '1.2.2'
__author__ = 'Jack Featherstone'
__credits__ = 'North Carolina State University; Okinawa Institute of Science and Technology' 



"""
Image loading and light gradient correction.
"""
import numpy as np

from PIL import Image
import cv2
from scipy.signal import savgol_filter

import matplotlib.pyplot as plt

IMAGE_TYPES = ['jpg',
               'jpeg',
               'png',
               'bmp',
               'tiff',
               'nef']

def imageType(file):
    """
    Returns the image type of a file.

    More or less just extracts the extension, but will return
    `None` if the extension is not that of an image, so can be
    used in the following way:

        listOfFiles = ['a.png', 'b.bmp', 'c.txt', ...]
        imageFiles = [f for f in listOfFiles if imageType(f)]

    Parameters
    ----------

    file : str or Path

    Returns
    -------

    imageExt : str or None
        Image extension if file is an image, else None.
    """
    isImage = file.lower().endswith(tuple(IMAGE_TYPES))
    return file.split('.')[-1] if isImage else None

def checkImageType(frame):
    """
    Make sure that the image is a proper image, and not a path

    Parameters
    ----------

    frame : str or numpy.ndarray
        Either a path to an image, or an image array

    Returns
    -------

    numpy.ndarray : The image array

    """
    if isinstance(frame, str):
        # I don't want to overwrite the image itself, so create a new var for that
        newFrame = np.array(Image.open(frame), dtype=np.uint8)
    else:
        newFrame = frame

    return newFrame


def lightCorrectionDiff(calibrationImage, verticalMask=None, horizontalMask=None, smoothCorrection=True, debug=False, smoothingKernel=31, channel=0, rectify=False):
    """
    
    """

    calImg = checkImageType(calibrationImage)

    # Slightly different behavior depending on whether we are passed a multichannel
    # image vs a grayscale one. For multichannel, we calculate a correction for each
    # channel separately.
    if calImg.ndim == 3:
        imgSize = calImg.shape[:2]
        numChannels = calImg.shape[-1]
    else:
        imgSize = calImg.shape
        numChannels = 1
        # Add a third dim, so we can treat multi/single channel images
        # exactly the same way
        calImg = calImg[:,:,None]

    if verticalMask is None:
        verticalMask = np.ones(imgSize)

    if horizontalMask is None:
        horizontalMask = np.ones(imgSize)

    fullLightCorrection = np.zeros((*imgSize, numChannels))

    for i in range(numChannels):

        verticallyMasked = calImg[:,:,i] * verticalMask
        horizontallyMasked = calImg[:,:,i] * horizontalMask

        # If there are no non-zero pixels, we just move on to the next channel
        # (and leave this correction as an array of zeros)
        if len(np.where(verticallyMasked != 0)[0]) == 0:
            continue

        # This excludes all values of zero, so that we get an actual pixel value we can directly add
        brightnessByRow = np.nanmean(np.where(verticallyMasked != 0, verticallyMasked, np.nan), axis=1)
        brightnessByColumn = np.nanmean(np.where(horizontallyMasked != 0, horizontallyMasked, np.nan), axis=0)

        # Now smooth the two curves
        # Can't say I know much about this filter, but it seems to work pretty well
        if smoothCorrection:
            smoothedBrightnessByColumn = savgol_filter(brightnessByColumn, smoothingKernel, 1)
            smoothedBrightnessByRow = savgol_filter(brightnessByRow, smoothingKernel, 1)
        else:
            smoothedBrightnessByColumn = brightnessByColumn
            smoothedBrightnessByRow = brightnessByRow

        # Now calculate the correction
        horizontalCorrection = np.mean(smoothedBrightnessByColumn) - smoothedBrightnessByColumn
        verticalCorrection = np.mean(smoothedBrightnessByRow) - smoothedBrightnessByRow

        # This object will have the same size as the image, and can just added to
        # any similar image to correct detected light gradients
        fullLightCorrection[:,:,i] = np.add.outer(verticalCorrection, horizontalCorrection)

        if rectify:
            fullLightCorrection[:,:,i] -= np.min(fullLightCorrection[:,:,i])

    # If we have a single channel image originally, we want to keep the same shape
    # for our return value -- so that the return can immediately be multipled by the
    # original image -- so we remove the last channel dimension
    if numChannels == 1:
        fullLightCorrection = fullLightCorrection[:,:,0]

    if debug:
        if numChannels > 1:
            fig, ax = plt.subplots(2, 3, figsize=(12, 8))

            channelNames = ['Red', 'Green', 'Blue']
            
            for i in range(3):
                ax[0,i].imshow(calImg[:,:,i])
                ax[0,i].set_title(f'Original {channelNames[i]} Channel')
                ax[1,i].imshow(calImg[:,:,i] + fullLightCorrection[:,:,i])
                ax[1,i].set_title(f'Corrected {channelNames[i]} Channel')

        else:
            fig, ax = plt.subplots(1, 2, figsize=(8,4))

            ax[0].imshow(calImg[:,:,0])
            ax[0].set_title('Original Image')
            ax[1].imshow(calImg[:,:,0] + fullLightCorrection)
            ax[1].set_title('Corrected Image')

        fig.tight_layout()
        plt.show()

    return fullLightCorrection.astype(calibrationImage.dtype)



"""
Methods to gauge how well force balance is satisfied for an ensemble,
and to convert between polar and cartesian systems.
"""

import numpy as np

import numba

def polarToCartesian(force, alpha, beta, collapse=True):
    """
    Convert a set of forces defined in polar coordinates (f, a, b),
    to cartesian coordinates (f_y, f_x).

    Parameters
    ----------

    force : float or np.ndarray[F] or list[F]
        The force magnitude, or an array/list of F force magnitudes.

    alpha : float or np.ndarray[F] or list[F]
        The alpha angle, or an array/list of F alpha angles.

    beta : float or np.ndarray[F] or list[F]
        The beta angle, or an array/list of F beta angles.

    collapse : bool
        Whether to collapse the force index dimension in the case that
        only a single force is provided.

    Returns
    -------

    forceArr : np.ndarray[F,2]
        An array of the cartesian components (y,x) of the forces. 

        If only a single force is provided (ie. `force`, `alpha` and `beta` are all
        floats) the first dimension will be omitted, leaving just `[f_y, f_x]`. See
        `collapse` for more information.
    """
    # Check to see if we were given multiple forces, or just a single one
    if hasattr(force, '__iter__'):
        forceArr = np.array(force)
        alphaArr = np.array(alpha)
        betaArr = np.array(beta)
        singleForce = False
    else:
        forceArr = np.array([force])
        alphaArr = np.array([alpha])
        betaArr = np.array([beta])
        singleForce = True
    
    cartesianForceArr = np.zeros((forceArr.shape[0], 2))
    for i in range(cartesianForceArr.shape[0]):
        # Note that this expression is not exactly the same as in K. E. Daniels et al.
        # Rev. Sci. Inst. 88 (2017). There is an extra negative on the alphas, since mine
        # appear to be defined backwards.
        # F_y
        cartesianForceArr[i,0] = forceArr[i] * np.cos(-alphaArr[i] + betaArr[i]) #(np.cos(betaArr[i,j]) * np.cos(alphaArr[i,j]) + np.sin(betaArr[i,j]) * np.sin(alphaArr[i,j]))
        # F_x
        cartesianForceArr[i,1] = -forceArr[i] * np.sin(-alphaArr[i] + betaArr[i]) #(-np.sin(betaArr[i,j]) * np.cos(alphaArr[i,j]) + np.cos(betaArr[i,j]) * np.sin(alphaArr[i,j]))
       

    # If we only have a single force, we should collapse that first dimension
    if singleForce and collapse:
        return cartesianForceArr[0]

    return cartesianForceArr


def testForceBalance(forceArr, alphaArr, betaArr, collapse=True):
    """
    Sum each of the cartesian force components to see how
    well an ensemble of forces satisfies force balance.

    Parameters
    ----------

    forceArr : np.ndarray[F] or np.ndarray[T,F]
        An array/list of F force magnitudes, possibly for T timesteps.

    alphaArr : np.ndarray[F] or np.ndarray[T,F]
        An array/list of F alpha angles, possibly for T timesteps.

    betaArr : np.ndarray[F] or np.ndarray[T,F]
        An array/list of F beta angles, possibly for T timesteps.

    collapse : bool
        Whether to collapse the timestep dimension in the case that
        only a single timestep is provided.

    Returns
    -------

    forceSumArr : np.ndarray[T,2]
        An array of the sum of each cartesian component (y,x) of the forces at each timestep.

        If only a single timestep is provided (ie. `forceArr`, `alphaArr` and `betaArr` are all
        1D arrays) the first dimension will be omitted, leaving just `[sum_f_y, sum_f_x]`. See
        `collapse` for more information.
    """
    # Check if we were given a single timestep, or multiple
    if len(np.shape(forceArr)) == 2:
        singleTimestep = False
        multiForceArr = np.array(forceArr)
        multiAlphaArr = np.array(alphaArr)
        multiBetaArr = np.array(betaArr)
    else:
        singleTimestep = True
        # TODO: Might need a transpose here
        multiForceArr = np.array([forceArr])
        multiAlphaArr = np.array([alphaArr])
        multiBetaArr = np.array([betaArr])

    forceSumArr = np.zeros((multiForceArr.shape[1], 2)) 
    # Sum up forces for each timestep
    for i in range(multiForceArr.shape[1]):
        cartForces = polarToCartesian(multiForceArr[:,i], multiAlphaArr[:,i], multiBetaArr[:,i], collapse=False)

        # sum_y
        forceSumArr[i,0] = np.sum(cartForces[:,0])
        # sum_x
        forceSumArr[i,1] = np.sum(cartForces[:,1])

    if singleTimestep and collapse:
        return forceSumArr[0]

    return forceSumArr


@numba.jit(nopython=True)
def singleParticleForceBalance(forceArr, alphaArr, betaArr):
    """
    **Does not currently work! Any calls to this function will just return the original
    arrays**

    Takes a set of forces acting on a single particle and ensures they obey
    force balance.

    The majority of this method is transpiled directly from Jonathan Kollmer's
    implementation:
    https://github.com/jekollmer/PEGS

    Parameters
    ----------

    forceArr : np.ndarray[N]
        Array of force magnitudes at each contact point.

    alphaArr : np.ndarray[N]
        Array of angles that define the direction of force at each contact point

    betaArr : np.ndarray[N]
        Array of angles that define the contact point of the forces, and therefore are
        not adjusted in the force balancing process

    Returns
    -------

    np.ndarray[N] : Magnitude of balanced forces

    np.ndarray[N] : Balanced contact angles alpha

    """

    # TODO: Get this function working
    print("Warning: force balance is not yet implemented, do not call the singleParticleForceBalance function!")
    return forceArr, alphaArr

    # Number of contacts (coordination number, often denoted by z)
    numContacts = len(forceArr)

    if numContacts < 2:
        # Can't do anything with only a single force
        return forceArr, alphaArr
    elif numContacts == 2:
        # For 2 forces, there is a unique process

        # The two force magnitudes must be equal
        balancedForceArr = np.array([forceArr[0], forceArr[0]])

        balancedAlphaArr = np.zeros(2) 
        dBeta = (betaArr[0] - betaArr[1]) / 2
        balancedAlphaArr[0] = np.arccos(np.sin(dBeta))
        if balancedAlphaArr[0] > np.pi/2:
            balancedAlphaArr[0] = np.arccos(np.sin(-dBeta))
        
        # And the other angle must be the opposite
        balancedAlphaArr[1] = - balancedAlphaArr[0]

        return balancedForceArr, balancedAlphaArr

    elif numContacts > 2:
        # We solve any z>2 contacts the same way
        balancedForceArr = np.zeros_like(forceArr)
        balancedAlphaArr = np.zeros_like(alphaArr)

        # To calculate the new force magnitudes, we add up vertical and
        # horizontal components of the other forces
        for i in range(numContacts):
            # These initializations are to not count the case where j = i
            sum1 = -forceArr[i] * np.sin(alphaArr[i])
            sum2 = -forceArr[i] * np.cos(alphaArr[i])
            for j in range(numContacts):
                sum1 += forceArr[j] * np.sin(alphaArr[j] + betaArr[j] - betaArr[i])
                sum2 += forceArr[j] * np.cos(alphaArr[j] + betaArr[j] - betaArr[i])

            balancedForceArr[i] = np.sqrt(sum1**2 + sum2**2)

        # To calculate new alpha values, we 
        for i in range(numContacts):
            sum3 = -balancedForceArr[i] * np.sin(alphaArr[i])
            for j in range(numContacts):
                sum3 += balancedForceArr[j] * np.sin(alphaArr[j])

            balancedAlphaArr[i] = np.arcsin(-sum3/balancedForceArr[i])


        return balancedForceArr, balancedAlphaArr




"""
Peak finding in arbitrary dimensional spaces via persistent homology.
"""
import numpy as np

import numba
import matplotlib.pyplot as plt


def _iterNeighbors(p, tShape, neighborInclusion=1):
    """
    Find the indices of all neighbors of an `d`-dimensional point.

    Used primarily in `pepe.topology.multiFindPeaks()`.

    Parameters
    ----------

    p : tuple[d]
        A tuple of indices representing a point in $Z^d$.

    tShape : tuple[d]
        The size of the grid/tensor in each dimension, such that
        points outside of the domain can be removed.
    """
    d = len(p)
    # Convert to numpy array to do math (which can't be done on tuples)
    nP = np.array(p, dtype=np.int16)
    # This will create an array that looks like [-2, -1, 0, 1, 2], [-1, 0, 1], etc.
    # depending on the value of neighborInclusion
    neighborDirections = np.arange(-neighborInclusion, neighborInclusion+1, dtype=np.int16)
        
    # Now find the combinations based on the possible values
    # This piece of code is partially taken from the method
    # itertools.product.
    # More or less the equivalent of itertools.product(neighborDirections, repeat=d)
    result = [[]] 
    for pool in [neighborDirections]*d:
        result = [x+[y] for x in result for y in pool]
       
    # Result is just the difference vectors from our point to the neighbors
    # (eg. [-1, 0] or [1, 1]) so we need to add these to our original point.
    neighbors = [nP + np.array(n) for n in result]
    
    # This element will always be the original point because the ordering for arange is
    # consistent such that the difference vector of [0,0,0,...] will always be at the
    # center (even after permuting for arbitrary dimensions).
    del neighbors[len(neighbors)//2]
   
    # Now make sure all of the points are within the domain of the data.
    for n in neighbors:
        if not True in ((np.array(tShape) - n - 1) < 0) and not True in (n < 0):
            yield tuple(n)


def findPeaksMulti(data, neighborInclusion=1, minPeakPrevalence=None, normalizePrevalence=True):
    """
    Identify peaks in multi-dimensional data using persistent topology.

    Peak prevalence is calculated as the age of topological features,
    which is not necessarily the same as the actual height of the peaks.

    Originally adapted from [1] (and partially from [2]), but generalized for data with
    an arbitrary number of dimensions. For background on persistence, see: [3] for
    application of persistence to signal processing, or [4] for general domain knowledge
    (these are just my personal recommendations, probably plenty of other resources out there).

    May struggle to find peaks in noisier data if the fluctuations due to
    noise are large enough to create a spiky-looking profile. This is because
    such fluctuations do represent topological objects, even if they are not ones
    of any physical relevance; this method does not distinguish between the two.
    In this case, smoothing functions are recommended to be used on the data before
    it is passed to this method. A higher `neighborInclusion` can partially address this.

    Method is not able to be optimized using `numba` (mostly due to the use of tuples,
    which `numba` can't type as easily as a list or array. While this method can identify
    peaks in 1D or 2D data, it is recommended to instead use either `pepe.topology.findPeaks1D()`
    or `pepe.topology.findPeaks2D()`, as these *are* optimized.

    Parameters
    ----------

    data : np.ndarray[d]
        An array of data points within which to identify peaks in d-dimensional space.

    neighborInclusion : int
        Controls the number of neighbors on either side of a point to look
        at when deciding which peak a given point belongs to. 

        eg. In 1D, a value of `1` means that `[i-1, i+1]` are considered neighbors,
        a value of `2` means that `[i-2, i-1, i+1, i+1]` are considered neighbors, etc.

        May help to supress the effects of noise in data, but should not be treated
        as a substitute for smoothing data properly beforehand.

    minPeakPrevalence : float or None
        The minimum prevalence of a peak (maximum - minimum) that will be
        returned to the user. If `normalizePrevalence` is True, this should
        be in the range `[0, 1]`, representing a percentage of the full range
        the data occupies.

    normalizePrevalence : bool
        Whether to normalize the prevalences such that they represent the
        percent of the data range that is spanned by a peak (True) or not (False).
        For example, if normalized, a prevalence of .4 would mean that the peak
        spans 40% of the total range of the data.

        Note that this normalization happens before using the minPeakPrevalence
        kwarg to clip smaller peaks, and thus the minimum value should be
        specified as a percent (eg. .3 for 30%) is the normalizePrevalence
        kwarg is True.

    Returns
    -------

    peakPositions : np.ndarray[N,d]
        The indices of peaks in the provided data, sorted from most
        to least prevalent (persistent).

    peakPrevalences : np.ndarray[N]
        The prevalence of each peak in the provided data, or the persistence
        of the topological feature. If `normalizePrevalence` is True, this will
        be normalized to the domain `[0, 1]`; otherwise, these values are
        equivalent to the maximum data value minus the minimum data value evaluated
        over all points that identify with a given peak.

    References
    ----------

    [1] Huber, Stefan. Persistent Topology for Peak Detection. 
    <https://www.sthu.org/blog/13-perstopology-peakdetection/index.html>

    [2] Huber, Stefan. Topological peak detection in two-dimensional data.
    <https://www.sthu.org/code/codesnippets/imagepers.html>

    [3] Huber, S. (2021). Persistent Homology in Data Science. In P. Haber,
    T. Lampoltshammer, M. Mayr, & K. Plankensteiner (Eds.), Data Science – Analytics
    and Applications (pp. 81–88). Springer Fachmedien. <https://doi.org/10.1007/978-3-658-32182-6_13>

    [4] Edelsbrunner, H., & Harer, J. (2010). Computational topology: An introduction.
    Chapter 7: Persistence. p. 149-156. American Mathematical Society. ISBN: 978-0-8218-4925-5

    """ 
    
    # This array contains the indices of the peak that each point
    # belongs to (assuming it does belong to a peak)
    # We start it at -1 such that we can check if a point belongs to
    # a certain peak with peakMembership[i,j,...] >= 0
    peakMembership = np.zeros_like(data, dtype=np.int16) - 1
    peakBirthIndices = []

    # Go through each value in our data, starting from highest and ending
    # with lowest. This is important because we will be merging points into
    # groups as we go, and so starting with the highest values means that
    # points will almost never have to be overwritten.
    # This is a bit complex here because we could have any number of dimensions, but
    # the gist is that we sort a flattened version of the array from highest to lowest,
    # then turn those 1d indices into Nd indices, then pair the indices together.
    # Each element of this array will be a set of indices that can index a single
    # element of data. 
    # eg. data[sortedIndices[0]] will be the largest value (for any number of dimensions)
    sortedIndices = np.dstack(np.unravel_index(np.argsort(data.flatten())[::-1], data.shape))[0]
    # To be able to actually index data with an element, they need to all be tuples
    sortedIndices = [tuple(si) for si in sortedIndices]
   
    # I've avoided using the index i here since we are iterating over sets of indices,
    # not just a single number
    for si in sortedIndices:
        # See if any neighbors have been assigned to a peak
        assignedNeighbors = [n for n in _iterNeighbors(si, data.shape, neighborInclusion) if peakMembership[n] >= 0]

        # If there aren't any assigned neighbors yet, then we create a new
        # peak
        if len(assignedNeighbors) == 0:
            peakMembership[si] = len(peakBirthIndices)
            peakBirthIndices.append(si)

        # If only a single one has been assigned, or all of the assigned peaks have the
        # same membership, then this point joins that peak
        elif len(assignedNeighbors) == 1 or len(np.unique([(peakMembership[n] for n in assignedNeighbors)])) == 1:
            peakMembership[si] = peakMembership[assignedNeighbors[0]]

        # Otherwise, we have to resolve a conflict between multiple, in which the
        # oldest one gains the new point.
        else:      
            # Find which one is the oldest
            order = np.argsort(data[peakBirthIndices[(peakMembership[n] for n in assignedNeighbors)]])[::-1]

            # New point joins oldest peak
            peakMembership[si] = peakMembership[assignedNeighbors[order][0]]

    # This is the position of the proper crest for each peak
    peakPositions = peakBirthIndices

    # Calculate the prevalence of each peak as the height of the birth points minus the
    # height of the death point
    # We could do this using the arrays we've stored along the way, but it's easier just to take the
    # max/min heights directly from the data
    peakPrevalences = np.array([data[peakPositions[i]] - np.min(data[peakMembership == i]) for i in range(len(peakBirthIndices))])
    # Also note that I have made the decision here to normalize these prevalences by the total range
    # of the data, such that a prevalence of .6 means that the peak spans 60% of the range of the data.
    # This can be altered with the normalizePrevalence kwarg
    if normalizePrevalence:
        dataRange = np.max(data) - np.min(data)
        peakPrevalences /= dataRange

    # Cut off small peaks, if necessary
    if minPeakPrevalence is not None:
        peakPositions = np.array([peakPositions[i] for i in range(len(peakPositions)) if peakPrevalences[i] > minPeakPrevalence])
        peakPrevalences = np.array([peakPrevalences[i] for i in range(len(peakPrevalences)) if peakPrevalences[i] > minPeakPrevalence])

    # Sort the peaks by their prevalence
    #order = np.argsort(peakPrevalences)[::-1]
    #peakPositions = peakPositions[order]
    #peakPrevalences = peakPrevalences[order]

    return (peakPositions, peakPrevalences)


def findPeaks2D(data, minPeakPrevalence=None, normalizePrevalence=True):
    """
    Identify peaks in 2-dimensional data using persistent topology.

    Peak prevalence is calculated as the age of topological features,
    which is not necessarily the same as the actual height of the peaks.

    Originally adapted from [1] (and partially from [2]), but adjusted for data with
    an specifically 2 dimensions. For background on persistence, see: [3] for
    application of persistence to signal processing, or [4] for general domain knowledge
    (these are just my personal recommendations, probably plenty of other resources out there).

    May struggle to find peaks in noisier data if the fluctuations due to
    noise are large enough to create a spiky-looking profile. This is because
    such fluctuations do represent topological objects, even if they are not ones
    of any physical relevance; this method does not distinguish between the two.
    In this case, smoothing functions are recommended to be used on the data before
    it is passed to this method.

    Method is optimized using `numba`.

    Parameters
    ----------

    data : np.ndarray[d]
        An array of data points within which to identify peaks in d-dimensional space.

    minPeakPrevalence : float or None
        The minimum prevalence of a peak (maximum - minimum) that will be
        returned to the user. If `normalizePrevalence` is True, this should
        be in the range `[0, 1]`, representing a percentage of the full range
        the data occupies.

    normalizePrevalence : bool
        Whether to normalize the prevalences such that they represent the
        percent of the data range that is spanned by a peak (True) or not (False).
        For example, if normalized, a prevalence of .4 would mean that the peak
        spans 40% of the total range of the data.

        Note that this normalization happens before using the minPeakPrevalence
        kwarg to clip smaller peaks, and thus the minimum value should be
        specified as a percent (eg. .3 for 30%) is the normalizePrevalence
        kwarg is True.

    Returns
    -------

    peakPositions : np.ndarray[N,d]
        The indices of peaks in the provided data, sorted from most
        to least prevalent (persistent).

    peakPrevalences : np.ndarray[N]
        The prevalence of each peak in the provided data, or the persistence
        of the topological feature. If `normalizePrevalence` is True, this will
        be normalized to the domain `[0, 1]`; otherwise, these values are
        equivalent to the maximum data value minus the minimum data value evaluated
        over all points that identify with a given peak.

    References
    ----------

    [1] Huber, Stefan. Persistent Topology for Peak Detection. 
    <https://www.sthu.org/blog/13-perstopology-peakdetection/index.html>

    [2] Huber, Stefan. Topological peak detection in two-dimensional data.
    <https://www.sthu.org/code/codesnippets/imagepers.html>

    [3] Huber, S. (2021). Persistent Homology in Data Science. In P. Haber,
    T. Lampoltshammer, M. Mayr, & K. Plankensteiner (Eds.), Data Science – Analytics
    and Applications (pp. 81–88). Springer Fachmedien. <https://doi.org/10.1007/978-3-658-32182-6_13>

    [4] Edelsbrunner, H., & Harer, J. (2010). Computational topology: An introduction.
    Chapter 7: Persistence. p. 149-156. American Mathematical Society. ISBN: 978-0-8218-4925-5

    """ 
    # Go through each value in our data, starting from highest and ending
    # with lowest. This is important because we will be merging points into
    # groups as we go, and so starting with the highest values means that
    # points will almost never have to be overwritten.
    # This is a bit complex here because we could have any number of dimensions, but
    # the gist is that we sort a flattened version of the array from highest to lowest,
    # then turn those 1d indices into Nd indices, then pair the indices together.
    # Each element of this array will be a set of indices that can index a single
    # element of data. 
    # eg. data[sortedIndices[0]] will be the largest value (for any number of dimensions)
    sortedIndices = np.dstack(np.unravel_index(np.argsort(data.flatten())[::-1], data.shape))[0]
    # To be able to actually index data with an element, they need to all be tuples
    sortedIndices = [tuple(si) for si in sortedIndices]
  
    # Numba gets mad if we pass lists between methods, so we have to turn our list
    # into a numba (typed) list
    typedSortedIndices = numba.typed.List(sortedIndices)

    peakMembership, peakBirthIndices = _findPeaks2DIter(data, typedSortedIndices)

    # Calculate the prevalence of each peak as the height of the birth points minus the
    # height of the death point
    # We could do this using the arrays we've stored along the way, but it's easier just to take the
    # max/min heights directly from the data
    peakPrevalences = np.array([data[peakBirthIndices[i]] - np.min(data[peakMembership == i]) for i in range(len(peakBirthIndices))])
    # Also note that I have made the decision here to normalize these prevalences by the total range
    # of the data, such that a prevalence of .6 means that the peak spans 60% of the range of the data.
    # This can be altered with the normalizePrevalence kwarg
    if normalizePrevalence:
        dataRange = np.max(data) - np.min(data)
        peakPrevalences /= dataRange

    # Cut off small peaks, if necessary
    if minPeakPrevalence is not None:
        peakBirthIndices = [peakBirthIndices[i] for i in range(len(peakBirthIndices)) if peakPrevalences[i] > minPeakPrevalence]
        peakPrevalences = np.array([peakPrevalences[i] for i in range(len(peakPrevalences)) if peakPrevalences[i] > minPeakPrevalence])

    # Sort the peaks by their prevalence
    #order = np.argsort(peakPrevalences)[::-1]
    #peakPositions = peakPositions[order]
    #peakPrevalences = peakPrevalences[order]

    return (peakBirthIndices, peakPrevalences)


@numba.njit(cache=True)
def _findPeaks2DIter(data, sortedIndices):
    """
    Iterative part of 2D peak finding, optimized using `numba`.

    As usual with these types of methods, there may be some statements
    that generally would be written better/easier another way, but end up
    looking weird because of a particular `numba` requirement.

    Not meant to be used outside of `pepe.topology.findPeaks2D()`.
    """
    # This array contains the indices of the peak that each point
    # belongs to (assuming it does belong to a peak)
    # We start it at -1 such taht we can check if a point belongs to
    # a certain peak with peakMembership[i,j,...] >= 0
    peakMembership = np.zeros_like(data, dtype=np.int16) - 1
    peakBirthIndices = []

    # I've avoided using the index i here since we are iterating over sets of indices,
    # not just a single number
    for si in sortedIndices:
        # See if any neighbors have been assigned to a peak
        # No options for expanding which points are considered neighbors here
        # since the 8 surrounding points should be fine.
        assignedNeighbors = [(si[0]+i, si[1]+j) for i in [0, 1, -1] for j in [0, 1, -1]][1:]
        # Cut off points outside the domain, or that haven't been assigned yet
        assignedNeighbors = [n for n in assignedNeighbors if n[0] >= 0 and n[1] >= 0 and n[0] < data.shape[0] and n[1] < data.shape[1] and peakMembership[n] >= 0]

        # If there aren't any assigned neighbors yet, then we create a new
        # peak
        if len(assignedNeighbors) == 0:
            peakMembership[si] = len(peakBirthIndices)
            peakBirthIndices.append(si)

        # If only a single one has been assigned, or all of the assigned peaks have the
        # same membership, then this point joins that peak
        elif len(assignedNeighbors) == 1 or len(np.unique(np.array([peakMembership[n] for n in assignedNeighbors]))) == 1:
            peakMembership[si] = peakMembership[assignedNeighbors[0]]

        # Otherwise, we have to resolve a conflict between multiple, in which the
        # oldest one gains the new point.
        else:      
            # Find which one is the oldest
            order = np.argsort(np.array([data[peakBirthIndices[peakMembership[n]]] for n in assignedNeighbors]))[::-1]
            # New point joins oldest peak
            peakMembership[si] = peakMembership[assignedNeighbors[order[0]]]

    return peakMembership, peakBirthIndices


@numba.njit(cache=True)
def findPeaks1D(data, minPeakPrevalence=None, normalizePrevalence=True):
    """
    Find peaks in one dimensional data using persistent homology.

    Peak prevalence is calculated as the age of topological features,
    which is not necessarily the same as the actual height of the peaks.

    Originally adapted from [1], with changes mostly pertaining to readability and 
    stronger emphasis on efficiency. For background on persistence, see: [2] for
    application of persistence to signal processing, or [3] for general domain knowledge.

    May struggle to find peaks in noisier data if the fluctuations due to
    noise are large enough to create a spiky-looking profile. This is because
    such fluctuations do represent topological objects, even if they are not ones
    of any physical relevance; this method does not distinguish between the two.
    In this case, smoothing functions are recommended to be used on the data before
    it is passed to this method.

    Parameters
    ----------

    data : np.ndarray
        An array of data points within which to identify peaks.

    minPeakPrevalence : float or None
        The minimum prevalence of a peak (maximum - minimum) that will be
        returned to the user. If `normalizePrevalence` is True, this should
        be in the range `[0, 1]`, representing a percentage of the full range
        the data occupies.

    normalizePrevalence : bool
        Whether to normalize the prevalences such that they represent the
        percent of the data range that is spanned by a peak (True) or not (False).
        For example, if normalized, a prevalence of .4 would mean that the peak
        spans 40% of the total range of the data.

        Note that this normalization happens before using the minPeakPrevalence
        kwarg to clip smaller peaks, and thus the minimum value should be
        specified as a percent (eg. .3 for 30%) is the normalizePrevalence
        kwarg is True.

    Returns
    -------

    peakPositions : np.ndarray[N]
        The indices of peaks in the provided data, sorted from most
        to least prevalent (persistent). If `returnValues` is True, this
        will instead be the values at these indices, or `data[peakPositions]`.

    peakPrevalences : np.ndarray[N]
        The prevalence of each peak in the provided data, or the persistence
        of the topological feature. If `normalizePrevalence` is True, this will
        be normalized to the domain `[0, 1]`; otherwise, these values are
        equivalent to the maximum data value minus the minimum data value evaluated
        over all points that identify with a given peak.

    References
    ----------

    [1] Huber, Stefan. Persistent Topology for Peak Detection. 
    <https://www.sthu.org/blog/13-perstopology-peakdetection/index.html>

    [2] Huber, S. (2021). Persistent Homology in Data Science. In P. Haber,
    T. Lampoltshammer, M. Mayr, & K. Plankensteiner (Eds.), Data Science – Analytics
    and Applications (pp. 81–88). Springer Fachmedien. <https://doi.org/10.1007/978-3-658-32182-6_13>

    [3] Edelsbrunner, H., & Harer, J. (2010). Computational topology: An introduction.
    Chapter 7: Persistence. p. 149-156. American Mathematical Society. ISBN: 978-0-8218-4925-5

    """

    # Entries will be a list(2) in this list, with the first element as the
    # left index, and the second element as the right index
    peakBoundsIndices = []
    peakBirthIndices = []
    # We don't actually need to record when a peak dies, since we will eventually
    # extract the persistence from the minimum value in each set of points,
    #peakDeathIndices = []

    # This array contains the indices of the peak that each point
    # belongs to (assuming it does belong to a peak)
    # We start it at -1 such taht we can check if a point belongs to
    # a certain peak with peakMembership[i] >= 0
    peakMembership = np.zeros(len(data), dtype=np.int16) - 1

    # Go through each value in our data, starting from highest and ending
    # with lowest. This is important because we will be merging points into
    # groups as we go, and so starting with the highest values means that
    # points will almost never have to be overwritten.
    sortedIndices = np.argsort(data)[::-1] # Reverse it, so it's highest -> lowest

    # Iterate over every index
    # It is important to note that we are iterating over in the order described above,
    # and so any indices calculated (for births, deaths, etc.) do not need to be first
    # passed through the sorted index array.
    # Eg. data[sortedIndices[peakMembership == 1]] would be incorrect to get all data
    # points in the peak labelled 1. Instead, data[peakMembership == 1] can be directly
    # used.
    for i in sortedIndices:
        # These bools represent whether the points to the left and right
        # (in the original data, not the height-sorted data) have been
        # assigned to a peak yet.
        leftAssigned = (i > 0 and peakMembership[i-1] >= 0)
        rightAssigned = (i < len(data)-1 and peakMembership[i+1] >= 0)

        # If neither left or right have been processed, this point begins
        # a new peak (which may later be merged into another one)
        if not leftAssigned and not rightAssigned:
            # Assign this point to be included in a new peak
            peakMembership[i] = len(peakBoundsIndices)
            # Create a new peak, and record it's birth
            peakBoundsIndices.append([i,i])
            peakBirthIndices.append(i)
            # See note above about death
            #peakDeathIndices.append(None)

        # If either the left or right peak exists (but only one) we merge this
        # point with the existing one
        if leftAssigned and not rightAssigned:
            # Expand the peak bounds to include this new point
            # (1 index because the peak is expanding to the right)
            peakBoundsIndices[peakMembership[i-1]][1] += 1
            peakMembership[i] = peakMembership[i-1]


        if rightAssigned and not leftAssigned:
            # Expand the peak bounds to include this new point
            # (0 index because the peak is expanding to the left)
            peakBoundsIndices[peakMembership[i+1]][0] -= 1
            peakMembership[i] = peakMembership[i+1]

        # If both left and right belong to peaks, that means one of them will
        # have to die, and the new point + the one who's group died get added
        # to the surviving group
        if rightAssigned and leftAssigned:
            # Determine which is taller
            if data[peakBirthIndices[peakMembership[i-1]]] > data[peakBirthIndices[peakMembership[i+1]]]:
                # If the left peak is higher, the right one dies
                # See note above about death
                #peakDeathIndices[peakMembership[i+1]] = i
                # Expand the bounds of the left peak to cover the right one
                # (1 index because the peak is expanding to the right)
                peakBoundsIndices[peakMembership[i-1]][1] = peakBoundsIndices[peakMembership[i+1]][1]
                # Save membership for current peak
                peakMembership[i] = peakMembership[i-1]
                # Save membership for right-most peak (that was just recently absorbed into this one).
                # You don't need to change every point in between, because those will never be
                # referenced again.
                peakMembership[peakBoundsIndices[peakMembership[i-1]][1]] = peakMembership[i-1]
            else:
                # If the right peak is higher, the left one dies
                # See note above about death
                #peakDeathIndices[peakMembership[i-1]] = i
                # Expand the bounds of the right peak to cover the left one
                # (0 index because the peak is expanding to the left)
                peakBoundsIndices[peakMembership[i+1]][0] = peakBoundsIndices[peakMembership[i-1]][0]
                # Save membership for current point
                peakMembership[i] = peakMembership[i+1]
                # Save membership for left-most point (that was just recently absorbed into this one).
                # You don't need to change every point in between, because those will never be
                # referenced again.
                peakMembership[peakBoundsIndices[peakMembership[i+1]][0]] = peakMembership[i+1]

    # This is the position of the proper crest for each peak
    peakPositions = np.array(peakBirthIndices)

    # Calculate the prevalence of each peak as the height of the birth points minus the
    # height of the death point
    # We could do this using the arrays we've stored along the way, but it's easier just to take the
    # max/min heights directly from the data
    peakPrevalences = np.array([data[peakPositions[i]] - np.min(data[peakMembership == i]) for i in range(len(peakBoundsIndices))])

    # Also note that I have made the decision here to normalize these prevalences by the total range
    # of the data, such that a prevalence of .6 means that the peak spans 60% of the range of the data.
    # This can be altered with the normalizePrevalence kwarg
    if normalizePrevalence:
        dataRange = np.max(data) - np.min(data)
        peakPrevalences /= dataRange

    # Cut off small peaks, if necessary
    if minPeakPrevalence is not None:
        peakPositions = peakPositions[peakPrevalences > minPeakPrevalence]
        peakPrevalences = peakPrevalences[peakPrevalences > minPeakPrevalence]

    # Sort the peaks by their prevalence
    order = np.argsort(peakPrevalences)[::-1]
    peakPositions = peakPositions[order]
    peakPrevalences = peakPrevalences[order]

    return (peakPositions, peakPrevalences)




"""
Coordinate transformation from discrete cartesian space to discrete polar space.
"""
import numpy as np

# I honestly don't think this will ever be used, but can't hurt to
# keep it around
def cartesianToPolar(mat, center=None, fillValue=np.nan):
    """
    Converts a discretized matrix in cartesian coordinates to polar
    coordinates.
    """
    
    if center is None:
        centerIndex = np.array([(mat.shape[0] - 1)/2, (mat.shape[1] - 1)/2])
        
    # First, we have to calculate the radii and angles    
    yArr = np.float64(np.arange(mat.shape[0])) - centerIndex[0]
    xArr = np.float64(np.arange(mat.shape[1])) - centerIndex[1]
    
    radiusArr = np.add.outer(yArr**2, xArr**2)
    angleArr = np.arctan2.outer(yArr, xArr)
    
    # Now we have to discretize these values, since they will become our new axes
    
    # Calculate what the smallest change in radius is
    dR = np.sort(np.unique(radiusArr.flatten()))[:2]
    dR = dR[1] - dR[0]
    print(dR)
    
    radiusIndices = np.int64(radiusArr / dR)
    
    # Calculate what the smallest change in angle we could possible have is
    # For this, we assume the farthest point from the center is in the bottom right
    # corner
    dTheta = abs(angleArr[-1,-1] - angleArr[-1,-2])
    print(dTheta)
    # Now make sure we are on the domain [0, 2pi] instead of [-pi, pi]
    angleArr += np.pi
    angleIndices = np.int64(angleArr / dTheta)

    minRI = np.min(radiusIndices.flatten())
    minThetaI = np.min(angleIndices.flatten())
    
    # Build the array, which may not be the same size as the previous array
    polarMat = np.zeros((np.max(radiusIndices.flatten()) - minRI + 1,
                         np.max(angleIndices.flatten()) - minThetaI + 1))
    
    polarMat[:,:] = fillValue
    
    for i in range(mat.shape[0]):
        for j in range(mat.shape[1]):
            polarMat[radiusIndices[i,j] - minRI, angleIndices[i,j] - minThetaI] = mat[i,j]
            
    return polarMat, (dR, dTheta)



"""
Outer products defined in a `numba`-friendly way.
"""
import numpy as np

import numba

@numba.njit(parallel=True, fastmath=True, cache=True)
def outerSubtract(x, y):
    """
    Perform outer subtraction of two arrays, x - y.

    Created because `numpy.subtract.outer` is not compatable with numba.

    Parameters
    ----------
    x : np.ndarray[N]
        First array.

    y : np.ndarray[M]
        Second array.

    Returns
    -------
    outer : np.ndarray[N,M]
        Matrix where the element [i,j] is x[i]-y[j].
    """
    outer = np.zeros((len(x), len(y)))
    for i in range(outer.shape[0]):
        for j in range(outer.shape[1]):
            outer[i,j] = x[i] - y[j]

    return outer




"""
Particle position tracking via Hough transform.
"""
import numpy as np

import cv2
from scipy.ndimage import filters

import numba

from pepe.analysis import gSquared
from pepe.preprocess import sobelEdgeDetection, laplacianEdgeDetection, cannyEdgeDetection

import matplotlib.pyplot as plt

def houghCircle(singleChannelFrame, radius, edgeDetection='laplacian', blurKernel=None, cannyEdgeThreshold=70, accumulatorThreshold=20, radiusTolerance=15, minSeparation=None, debug=False):
    """
    Perform Hough circle detection on the provided image.

    For background on the Hough method (especially in the context
    of granular materials), see [1].

    OpenCV's implementation of the  Hough transform is used, so this
    method can be considered a wrapper for `cv2.HoughCircle()`.

    Tends to be less accurate than the convolution method (see
    `pepe.tracking.convCircle()`) but usually offers shorter computation
    times, especially when the radius is not well known. For cases
    in which the radius is known to a tolerance of ~10 pixels, the convolution
    method is recommended.

    Parameters
    ----------

    singleChannelFrame : np.uint8[H, W]
        The frame to perform circle detection on. Should not include multiple
        color channels.

    radius : float or [float, float]
        The approximate radius to look for, or a range of radii to look between.
        If a single value is provided, a range will be constructed around that
        value with the `radiusTolerance` parameter.

    edgeDetection : ['sobel', 'laplacian', 'canny', None]
        What type of edge detection should be performed on the image before the hough
        algorithm is applied.

    cannyEdgeThreshold : float
        Value between 0 and 100 representing the threshold for edge detection. Smaller
        values means the edges of circles do not have to be as strongly defined, but
        this also means you may get false positive detections.

    accumulatorThreshold : float
        Value between 0 and 100 representing the threshold for the accumulator (how
        the Hough method votes on which objects are circles). Higher values should
        be used when circles are well defined, and lower values for weaker signals.

    radiusTolerance : float
        Value used to construct a radius interval assuming only a single value is
        passed for parameter `radius`. Interval will be [radius - radiusTolerance,
        radius + radiusTolerance].

    minSeparation : float
        The minimum distance allowed between detected circles. If None (the
        default value) this will be calculated as twice the radius minus the
        radius tolerance.

    debug : bool
        Whether or not to draw debug information on a plot.

    Returns
    -------

    centers : np.ndarray[N,2]
        The detected circles' centers.

    radii : np.ndarray[N]
        The detected circles' radii.

    References
    ----------

    [1] Franklin, S. V., & Shattuck, M. D. (Eds.). (2016). Handbook
    of Granular Materials. Chapter 2: Experimental Techniques. p.68-70.
    CRC Press. https://doi.org/10.1201/b19291

    """

    # Check if we are given a range of radii or just a single value
    minRadius = None
    maxRadius = None

    if hasattr(radius, '__iter__'):
        if len(radius) == 2:
           minRadius = radius[0]
           maxRadius = radius[1]
    else:
        # Use the tolerance
        minRadius = radius - radiusTolerance
        maxRadius = radius + radiusTolerance


    if minRadius is None or maxRadius is None:
        raise Exception('Exception parsing expected radius!')

    if blurKernel is not None:
        blurredImage = cv2.blur(singleChannelFrame, (blurKernel, blurKernel))
    else:
        blurredImage = singleChannelFrame

    if edgeDetection == 'laplacian':
        detectImage = laplacianEdgeDetection(blurredImage)
    elif edgeDetection == 'sobel':
        detectImage = sobelEdgeDetection(blurredImage)
    elif edgeDetection == 'canny':
        detectImage = cannyEdgeDetection(blurredImage)
    else:
        detectImage = blurredImage


    # We don't expect too much overlap, so this should be a reasonable
    # separation distance
    if minSeparation is None:
        minSeparation = 2*minRadius - radiusTolerance

    detectedCircles = cv2.HoughCircles(np.array(detectImage, dtype=np.uint8), 
                    cv2.HOUGH_GRADIENT,
                    dp=1, # Accumulator (no idea what it does, but 1 is good)
                    minDist=minSeparation, # Minimum distance between circles
                    param1=cannyEdgeThreshold, # Threshold for canny edge detection
                    param2=accumulatorThreshold, # Accumulator threshold
                    minRadius=minRadius,
                    maxRadius=maxRadius)

    if detectedCircles is None:
        return np.array([]), np.array([])

    centers = np.array(list(zip(detectedCircles[0,:,1], detectedCircles[0,:,0])))
    radii = detectedCircles[0,:,2]


    if debug:
        fig, ax = plt.subplots(1, 3, figsize=(10,4))
        ax[0].imshow(singleChannelFrame)
        ax[0].set_title('Original image')

        ax[1].imshow(detectImage)
        ax[1].set_title('Detection image')

        ax[2].imshow(detectImage)
        ax[2].set_title('Detected circles')
        for i in range(len(centers)):
            c = plt.Circle(centers[i][::-1], radii[i], color='red', fill=False, linewidth=1)
            ax[2].add_artist(c)

        fig.tight_layout()

    return centers, radii







"""
Start from a directory of images and solve for all of the particle positions, orientations, and forces.
"""

import numpy as np
import os
import cv2
import time
import pickle
import inspect
import ast
import tqdm

import matplotlib.pyplot as plt

from PIL import Image, ImageDraw
from IPython.display import clear_output

import pepe
from pepe.preprocess import checkImageType, lightCorrectionDiff, circularMask
from pepe.analysis import initialForceSolve, forceOptimize, gSquared, g2ForceCalibration, singleParticleForceBalance, forceOptimizeArgDTypes
from pepe.tracking import houghCircle, convCircle, angularConvolution, circleTrackArgDTypes
from pepe.simulate import genSyntheticResponse
from pepe.utils import preserveOrderArgsort, rectangularizeForceArrays, explicitKwargs, parseList
from pepe.visualize import genColors, visCircles, visForces, visContacts, visRotation
from pepe.topology import findPeaksMulti

# All of the dtypes of the args for the below method
# The following args are not included, because they are not
# important: progressBarOffset, progressBarTitle
forceSolveArgDTypes  = {"imageDirectory": str,
            "imageExtension": str,
            "imageEndIndex": int,
            "imageStartIndex": int,
            "carryOverAlpha": bool,
            "carryOverForce": bool,
            "showProgressBar": bool,
            "lightCorrectionImage": str,
            "lightCorrectionVerticalMask": str,
            "lightCorrectionHorizontalMask": str,
            "g2CalibrationImage": str,
            "g2CalibrationCutoffFactor": float,
            "maskImage": str,
            "cropXMin": int,
            "cropXMax": int,
            "circleDetectionMethod": str,
            "guessRadius": float,
            "fSigma": float,
            "pxPerMeter": float,
            "brightfield": bool,
            "contactPadding": int,
            "g2MaskPadding": int,
            "contactMaskRadius": int,
            "peBlurKernel": int,
            "requireForceBalance": bool,
            "circleTrackingChannel": int,
            "circleTrackingGain": float,
            "circleTrackingKwargs": dict,
            "photoelasticChannel": int,
            "photoelasticGain": float,
            "optimizationKwargs": dict,
            "maxBetaDisplacement": float,
            "forceNoiseWidth": float,
            "alphaNoiseWidth": float,
            "saveMovie": bool,
            "pickleArrays": bool,
            "outputRootFolder": str,
            "outputExtension": str,
            "genFitReport": bool,
            "performOptimization": bool,
            "inputSettingsFile": str,
            "debug": bool}

# These dtypes don't matter, so we don't need to carry them
# through the settings file, but we want to explicitly define
# them so that the test function in pepe/test/test_auto.py
# doesn't catch them as missing
excludedArgs = ["progressBarOffset", "progressBarTitle"]

# Decorator that allows us to identify which keyword arguments were explicitly
# passed to the function, and which were left as default values. See beginning
# of method code for more information/motivation.
@explicitKwargs()
def forceSolve(imageDirectory, guessRadius=0.0, fSigma=0.0, pxPerMeter=0.0, brightfield=True, contactPadding=15, g2MaskPadding=2, contactMaskRadius=30, lightCorrectionImage=None, lightCorrectionHorizontalMask=None, lightCorrectionVerticalMask=None, g2CalibrationImage=None, g2CalibrationCutoffFactor=.9, maskImage=None, cropXMin=None, cropXMax=None, peBlurKernel=3, imageExtension='bmp', requireForceBalance=False, imageStartIndex=None, imageEndIndex=None, carryOverAlpha=True, carryOverForce=True, circleDetectionMethod='convolution', circleTrackingKwargs={}, circleTrackingChannel=0, circleTrackingGain=1., maxBetaDisplacement=.5, photoelasticChannel=1, photoelasticGain=1., forceNoiseWidth=.03, alphaNoiseWidth=.01, optimizationKwargs={}, performOptimization=True, debug=False, showProgressBar=True, progressBarOffset=0, progressBarTitle=None, saveMovie=False, outputRootFolder='./', inputSettingsFile=None, pickleArrays=True, genFitReport=True, outputExtension=''):
    """
    Complete pipeline to solve for forces and particle positions for all image files
    in a directory. Results will be returned and potentially written to various files.
    See `Returns` section for more information

    Expects all particles to be the same (or very similar) sizes. This assumption is made
    by the calculation of the gradient squared calibration value, which is computed just
    once using the guess of the radii. This should not be a problem if the radii are only
    slightly varied (~10 pixels or something) but any more than that and errors will begin
    to accumulate.

    This method has **a lot** of arguments; it is intended to be used once reasonable
    values for all of these have already been found. While the `debug` option for this function
    is very helpful, it is recommended to utilize the various notebooks/examples to find good
    choices for parameters first.

    The output readme file can also serve as a cache of the parameter values/settings, which
    can be passed back to future calls of this method using the `inputSettingsFile` argument.

    Parameters
    ----------

    imageDirectory : str
        An absolute or relative path to the directory containing the images that are
        to be analyzed. The names of the images need not follow any particular naming
        scheme, but they should be such that sorting the list alphabetically will give
        the images in their proper order.

    guessRadius : float
        The radius of the particles to be detected, in pixels. As of now, this value will
        be taken as the particle radius, but future versions may be able to vary this to
        find the optimal value.

        Currently no support for particles of different sizes.

    fSigma : float
        Stress optic coefficient, relating to material thickness, wavelength of light and
        other material property (denoted as C in most literature; sometimes also called the
        "stress optic coefficient").

    pxPerMeter : float
        The number of pixels per meter in the images. Depends on the camera, lens, and 
        zoom settings used to capture the images.

        Note that this is **not** the inverse of the number of pixels per meter, as is used
        in some of the other force solving implementations.

    brightfield : bool
        Whether the images are captured using a brightfield polariscope (`True`) or
        a darkfield polariscope (`False`).

    contactPadding : int
        Maximum distance (in pixels) between a particle's edge and the wall or the edges of two
        particles that will still be considered a potential force-bearing contact.

    g2MaskPadding : int or float
        Number of pixels to ignore at the edge of each particle when calculating the average G^2.
        If float value < 1 is passed, gradient mask radius will be taken as that percent of the full
        particle radius. A value of 0 means no padding is included.

    contactMaskRadius : float
        The radius of the circular mask that will be constructed around each contact to estimate
        the magnitude of the force using the gradient squared in that region.

    lightCorrectionImage : str or np.ndarray[H,W[,C]]
        The path to an image (or an array representing an image) that contains no particles, and
        can be used to correct for any light gradients present in the actual data.

    lightCorrectionHorizontalMask : str or np.ndarray[H,W[,C]]
        A mask array, containing values of `0` or `1`, with the latter representing areas over which
        the horizontal light correction should be calculated. Can also be a path to an image.

    lightCorrectionVerticalMask : str or np.ndarray[H,W[,C]]
        A mask array, containing values of `0` or `1`, with the latter representing areas over which
        the vertical light correction should be calculated. Can also be a path to an image.

    g2CalibrationImage : str or np.ndarray[H,W,C]
        An image with a single particle (or at least one particle) that has no force acting on it.
        Used to determine the base level of gradient squared (due to noise) for a free particle. Can
        also be a path to an image.

    g2CalibrationCutoffFactor : float
        The factor that is multipled by the mean gradient squared value of the particles in the 
        calibration image. Any particle that has an average gradient squared value below the 
        calibration value multipled by this factor will be assumed to have no forces acting on it.

    maskImage : str or np.ndarray[H,W,C]
        A mask array, containing values of `0` or `1`, with the latter representing the regions of
        importance for the image. Used in detecting particles, generating initial guesses, and
        calculating error during non-linear optimization. can also be a path to an image.

    cropXMin : int or None
        Left bound to crop down the image in the x direction.

    cropXMax : int or None
        Right bound to crop down the image in the x direction.

    peBlurKernel : int
        The kernel size that will be used for bluring the photoelastic channel of each image, to
        reduce noise. Should be an odd integer.

    imageExtension : str
        The extension of the image files that will be read in from `imageDirectory`. Should not include
        the '.' before the extension.

    requireForceBalance : bool
        Whether to impose particle-wise force balance at each step (`True`) or to take the results of
        the optimization process as they are (`False`).

        Currently WIP, and does not do anything.

    forceBalanceWeighting : float
        If a non-zero positive value, adds a contribution to the optimization cost
        pertaining to how well the ensemble of forces satisfy force balance.

    imageStartIndex : int or None
        The index of which image to start at when analyzing the files in `imageDirectory`. Value
        of `None` will start at the first (alphabetically sorted) image.

    imageEndIndex : int or None
        The index of which image to end at when analyzing the files in `imageDirectory`. Value
        of `None` will end at the last (alphabetically sorted) image.

    circleDetectionMethod : ['convolution' or 'hough']
        Whether to use the convolution or hough circle detection method to identify particles.

        See `pepe.tracking.convCircle()` and `pepe.tracking.houghCircle()` for more information.

    circleTrackingKwargs : **kwargs
        Keyword arguments to be passed to the selected circle tracking function.

        See `pepe.tracking.convCircle()` and `pepe.tracking.houghCircle()` for more information.

    circleTrackingChannel : int
        The channel of the image that will be used to track the particles. `0` for red, `1` for
        green, and `2` for blue.

    circleTrackingGain : float
        The gain applied to the intensity in the circle tracking channel.

    maxBetaDisplacement : float
        The maximum distance (angle) that a force can move between frames and still be identified
        as the same force. If a force moves more than this value, it will still be recorded as a force,
        but will be considered a new and independent force from any of the ones in the previous frame.

    photoelasticChannel : int
        The channel of the image that will be used to gauge the photoelastic response. `0` for red, `1` for
        green, and `2` for blue.

    photoelasticGain : float
        The gain applied to the intensity in the photoelastic channel.

    forceNoiseWidth : float or None
        The width of the gaussian distribution (centered at 0) that noise is sampled from to add to the
        force guesses (potentially from the previous frame). This is done to avoid getting stuck in a local
        minimum for too long (adds some Monte-Carlo-esque behavior to the solving).

    alphaNoiseWidth : float or None
        The width of the gaussian distribution (centered at 0) that noise is sampled from to add to the
        alpha guesses (potentially from the previous frame). This is done to avoid getting stuck in a local
        minimum for too long (adds some Monte-Carlo-esque behavior to the solving).

    optimizationKwargs : **kwargs
        Keyword arguments to be passed to the optimization process.

        For more information, see `pepe.analysis.forceOptimize()`.

    performOptimization : bool
        Whether or not to perform optimization on the particles.

        Mostly included as a debug option, but any real data analysis should
        utilize the optimization, as the initial guessing is often not nearly
        accurate enough to get any real results.

    debug : bool
        Whether to print progress updates for each frame to the screen (`True`) or not (`False`).

    showProgressBar : bool
        Whether to show a progress bar throughout the analysis (`True`) or not (`False`). Uses
        `tqdm` library.

    progressBarOffset : int
        The number of lines to offset the progress bar by. Generally an internal variable
        used when multiple threads are active.

    progressBarTitle : str
        The text to be written to the left of the progress bar. Generally an internal variable
        controlled by some solving script.

    saveMovie : bool
        Whether to save a compiled gif of the reconstructed forces at each frame at the end (`True`)
        or not (`False`).

    outputRootFolder : str
        The location where the output folder (potentially containg the movie, pickle files, readme, etc.)
        will be created. Output folder itself will be named after the `imageDirectory`, with '_Synthetic'
        appended to the end.

    pickleArrays : bool
        Whether to save the forces, betas, alphas, centers, and radii as pickle files (`True`) or not (`False`).
        Files will be located in the output folder (see `outputRootFolder`).

    inputSettingsFile : str
        Path to a readme file containg parameters for the solving process, likely generated from
        a previous iteration of the program. Explicitly passed arguments will override those that
        are included in the settings file.

        Currently WIP and does not do anything.

    genFitReport : bool
        Whether or not to generate a fit report of the results, including errors per frame,
        examinations of all particles/forces, and settings, compiled in a latex pdf.

        Will generate both the compiled file 'FitReport.pdf' and the source directory
        'FitReport_src/'.
    
    Returns
    -------

    rectForceArr : list[P](np.ndarray[F,T])
        A list of arrays representing the force magnitude for each force on each particle.

    rectAlphaArr : list[P](np.ndarray[F,T])
        A list of arrays representing the alpha angle for each force on each particle.

    rectBetaArr : list[P](np.ndarray[F,T])
        A list of arrays representing the beta angle for force on each particle.

    rectCenterArr : np.ndarray[P,T,2]
        Particle centers for each timestep. Elements take on a value of `[np.nan, np.nan]`
        if the particle does not exist for a given timestep.

    rectRadiusArr : np.ndarray[P,T]
        Particle radii for each timestep. Elements take on a value of `np.nan` if the particle
        does not exist for a given timestep.


    Depending on kwarg values, several files may be written created in the output
    folder, which will be located in `outputRootFolder` and named according
    to:  '<`imageDirectory`>_Synthetic/'.
    """
   
    overallStartTime = time.perf_counter()

    # For the sake of saving the options to a readme file (and potentially)
    # reading them back out, it is easiest to keep all of the settings in a
    # dictionary

    # We have 3 layers of precedence for reading in settings:
    # 1. Explicitly passed kwarg
    # 2. Read in from settings file
    # 3. Default value of a kwarg
    # So we assign the elements of our settings dict in opposite order

    # 3. All of the default values
    # The following variables are not present:
    # progressBarOffset, progressBarTitle
    # This is because we don't care about saving them
    settings = {"imageDirectory": os.path.abspath(imageDirectory) + '/', # Convert to absolute path
                "imageExtension": imageExtension,
                "imageEndIndex": imageEndIndex,
                "imageStartIndex": imageStartIndex,
                "carryOverAlpha": carryOverAlpha,
                "carryOverForce": carryOverForce,
                "lightCorrectionImage": lightCorrectionImage,
                "lightCorrectionVerticalMask": lightCorrectionVerticalMask,
                "lightCorrectionHorizontalMask": lightCorrectionHorizontalMask,
                "g2CalibrationImage": g2CalibrationImage,
                "g2CalibrationCutoffFactor": g2CalibrationCutoffFactor,
                "maskImage": maskImage,
                "cropXMin": cropXMin,
                "cropXMax": cropXMax,
                "circleDetectionMethod": circleDetectionMethod,
                "guessRadius": guessRadius,
                "fSigma": fSigma,
                "pxPerMeter": pxPerMeter,
                "brightfield": brightfield,
                "contactPadding": contactPadding,
                "g2MaskPadding": g2MaskPadding,
                "contactMaskRadius": contactMaskRadius,
                "peBlurKernel": peBlurKernel,
                "requireForceBalance": requireForceBalance,
                "circleTrackingChannel": circleTrackingChannel,
                "circleTrackingGain": circleTrackingGain,
                "photoelasticChannel": photoelasticChannel,
                "photoelasticGain": photoelasticGain,
                "maxBetaDisplacement": maxBetaDisplacement,
                "forceNoiseWidth": forceNoiseWidth,
                "alphaNoiseWidth": alphaNoiseWidth,
                "showProgressBar": showProgressBar,
                "saveMovie": saveMovie,
                "pickleArrays": pickleArrays,
                "outputRootFolder": outputRootFolder,
                "outputExtension": outputExtension,
                "genFitReport": genFitReport,
                "performOptimization": performOptimization,
                "debug": debug}

    # For the next step, we will need to know all of the data types of each
    # argument (to properly cast). Because certain arguments have None as a default
    # value, we can't automatically generate this information.
    # See above this method for the list of these, since they are also used
    # in the TrialObject file

    # We need to do the same thing for the kwargs for both
    # circle tracking and optimization

    # These have both been moved to either the tracking/DTypes.py
    # or the analysis/ForceSolve.py files, respectively

    # Now add all the dictionaries together
    argDTypes = forceSolveArgDTypes.copy()

    argDTypes.update(circleTrackArgDTypes)
    argDTypes.update(forceOptimizeArgDTypes)

    # 2. Anything read in from a settings file
    # Note that it works to our advantage that we already have values for most entries,
    # since the settings file doesn't include type information, so we need the old
    # values to cast properly. 1. is actually contained in here as well, because
    # we can just check to see if that variable was explicitly passed before overwriting it.

    # 1. The kwargs that are explicitly passed
    # This one is a little tricky, because there isn't a super great way by default
    # to differentiate whether a kwarg is explicitly passed or is its default value
    # (without just keeping a list of default values). I also don't want to
    # replace the entire function signature with (*args, **kwargs) because then the
    # documentation would not be as good (I think). So the solution here is to define
    # a decorator that has the (*args, **kwargs) signature, and to create an attribute
    # of this method that is a list of the kwargs that are explicitly passed to the
    # decorator. See `pepe.utils.explicitKwargs()` for more info.
    if inputSettingsFile is not None:
        if os.path.exists(inputSettingsFile):
            fileObj = open(inputSettingsFile, 'r')
            for line in fileObj:
                # Check each line and see if it looks like a dictionary value
                split = line.split(':')
                # Read settings into the master settings file
                if len(split) == 2 and split[0].strip() in argDTypes.keys() and not split[0].strip() in forceSolve.explicit_kwargs:
                    # Cast to the type of the value already in the dict
                    if split[1].strip() == 'None':
                        settings[split[0].strip()] = None
                    else:
                        if '[' in split[1]:
                            settings[split[0].strip()] = parseList(split[1].strip(), dtype=argDTypes[split[0].strip()])
                        else:
                            # Bools need a special condition
                            if argDTypes[split[0].strip()] is bool:
                                val = split[1].strip() == 'True'
                            else:
                                val = argDTypes[split[0].strip()](split[1].strip())

                            settings[split[0].strip()] = val

        else:
            print(f'Warning: provided settings file does not exist! Attempting to run regardless...')

    # While the following variables all have a default value of 0, they cannot actually
    # be left as this value. The reason they have a default value is so that if these
    # values are indicated by a settings file, we don't want to have to enter them again.
    # So here, we make sure we have values for them all, either explicitly passed or read in.
    requiredVars = ["guessRadius", "fSigma", "pxPerMeter"]
    for r in requiredVars:
        assert settings[r] != 0, f'Error: {r} value not supplied explicitly or implicitly!'

    # Now carry over the kwargs that are sent to the optimization procedure into that
    # dictionary. We can find the names of arguments by using the `inspect` library
    possibleOptimKwargs = list(inspect.signature(forceOptimize).parameters.keys())
    for pkw in possibleOptimKwargs:
        if pkw in settings.keys():
            optimizationKwargs[pkw] = settings[pkw]

    # We want to do the same thing for the circle tracking function, but we don't
    # yet know which circle tracking function we are using yet, so we'll carry
    # that over a bit later.

    # Find all images in the directory
    imageFiles = os.listdir(settings["imageDirectory"])

    # This goes before the sorting/extension filtering so we can get more specific
    # error messages (and we have another one of these below)
    if len(imageFiles) < 1:
        print(f'Error: directory {imageDirectory} contains no files!')
        return None

    imageFiles = np.sort([img for img in imageFiles if img[-len(settings["imageExtension"]):] == settings["imageExtension"]])

    # DEBUG:
    #imageFiles = imageFiles[::-1]

    # We have to do the end index first, so it doesn't mess up the start one
    if settings["imageEndIndex"] is not None:
        imageFiles = imageFiles[:min(settings["imageEndIndex"], len(imageFiles))]

    if settings["imageStartIndex"] is not None:
        imageFiles = imageFiles[max(settings["imageStartIndex"], 0):]

    # Make sure we still have some proper images
    if len(imageFiles) < 1:
        print(f'Error: directory \'{settings["imageDirectory"]}\' contains no files with extension \'{settings["imageExtension"]}\'!')
        return None

    xB = [settings["cropXMin"], settings["cropXMax"]]

    imageSize = checkImageType(settings["imageDirectory"] + imageFiles[0])[:,xB[0]:xB[1],0].shape

    # This will calculation the light correction across the images 
    if settings["lightCorrectionImage"] is not None:
        # Convert to absolute paths if they are paths
        if type(settings["lightCorrectionImage"]) is str:
            settings["lightCorrectionImage"] = os.path.abspath(settings["lightCorrectionImage"])
        if type(settings["lightCorrectionVerticalMask"]) is str:
            settings["lightCorrectionVerticalMask"] = os.path.abspath(settings["lightCorrectionVerticalMask"])
        if type(settings["lightCorrectionHorizontalMask"]) is str:
            settings["lightCorrectionHorizontalMask"] = os.path.abspath(settings["lightCorrectionHorizontalMask"])
            
        cImageProper = checkImageType(settings["lightCorrectionImage"])[:,xB[0]:xB[1]]
        vMask = checkImageType(settings["lightCorrectionVerticalMask"])[:,xB[0]:xB[1]]
        hMask = checkImageType(settings["lightCorrectionHorizontalMask"])[:,xB[0]:xB[1]]

        if vMask.ndim == 3:
            vMask = vMask[:,:,0]
        if hMask.ndim == 3:
            hMask = hMask[:,:,0]
        lightCorrection = lightCorrectionDiff(cImageProper, vMask, hMask)
        
        trackCorrection = lightCorrection[:,:,settings["circleTrackingChannel"]]
        peCorrection = lightCorrection[:,:,settings["photoelasticChannel"]]
    else:
        # It probably isn't great hygiene to have this variableflip between a single
        # value and an array, but you can always add a scalar to a numpy array, so
        # this is the easiest way (since we haven't loaded any images yet)
        trackCorrection = 0
        peCorrection = 0

    # Load up the mask image, which will be used to remove parts of the images
    # that we don't care about, and also potentially indicate which particles
    # are close to the boundary.
    if settings["maskImage"] is not None:
        maskArr = checkImageType(settings["maskImage"])[:,xB[0]:xB[1]]
        ignoreBoundary = False

    else:
        # Same deal as above: scalar multiplication functions exactly how we want
        # in the case that we don't have a mask, so it's just easier to do this.
        maskArr = 1
        ignoreBoundary = True

    # Which method we will be using to detect circles
    if settings["circleDetectionMethod"] == 'convolution':
        circFunc = convCircle
    elif settings["circleDetectionMethod"] == 'hough':
        circFunc = houghCircle
    else:
        print(f'Error: circle detection option \'{settings["circleDetectionMethod"]}\' not recognized!')
        return None

    # Now that we have a circle tracking function, we can carry over any possible kwargs
    possibleCircleKwargs = list(inspect.signature(circFunc).parameters.keys())
    for pkw in possibleCircleKwargs:
        if pkw in settings.keys():
            circleTrackingKwargs[pkw] = settings[pkw]

    # Calculate the lowest g2 value that we care about, so we can throw everything
    # that is below that away when solving (optional)
    checkMinG2 = False
    if settings["g2CalibrationImage"] is not None:
        g2CalImage = checkImageType(settings["g2CalibrationImage"])[:,xB[0]:xB[1]]

        g2CalPEImage = cv2.blur(settings["photoelasticGain"]*(g2CalImage[:,:,settings["photoelasticChannel"]] + peCorrection).astype(np.float64) / 255, (settings["peBlurKernel"],settings["peBlurKernel"]))
        # Locate particles
        centers, radii = circFunc(settings["circleTrackingGain"]*(g2CalImage[:,:,settings["circleTrackingChannel"]] + trackCorrection) * maskArr[:,:,0], settings["guessRadius"], **circleTrackingKwargs)
        # There should only be 1 particle in the calibration image
        if len(centers) < 0:
            print(f'Warning: Gradient-squared calibration image does not contain any particles! Ignoring...')
        else:
            particleMask = circularMask(g2CalPEImage.shape, centers[0], radii[0])[:,:,0]
            gSqr = gSquared(g2CalPEImage)
            minParticleG2 = np.sum(gSqr * particleMask) / np.sum(particleMask) * settings["g2CalibrationCutoffFactor"]
            checkMinG2 = True


    # The arrays that we will be building for each timestep. It is better to just
    # use an untyped list since the arrays are all triangular and whatnot.
    centersArr = []
    radiiArr = []
    
    forceArr = []
    betaArr = []
    alphaArr = []

    imageArr = []
    errorArr = []

    # For keeping track of time (though will only be display if debug=True)
    trackingTimes = np.zeros(len(imageFiles))
    initialGuessTimes = np.zeros(len(imageFiles))
    optimizationTimes = np.zeros(len(imageFiles))
    miscTimes = np.zeros(len(imageFiles))
    totalFailedParticles = 0

    errorMsgs = []

    if settings["showProgressBar"]:
        bar = tqdm.tqdm(total=len(imageFiles)+1, position=progressBarOffset, desc=progressBarTitle)

    # Calculate the gradient-squared-to-force calibration value
    g2Cal = g2ForceCalibration(settings["fSigma"], settings["guessRadius"], settings["pxPerMeter"])

    # The big loop that iterates over every image
    for i in range(len(imageFiles)):

        image = checkImageType(settings["imageDirectory"] + imageFiles[i])[:,xB[0]:xB[1]]
        # Convert to floats on the domain [0,1], so we can compare to the output of 
        # genSyntheticResponse()
        peImage = cv2.blur(settings["photoelasticGain"]*(image[:,:,settings["photoelasticChannel"]] + peCorrection).astype(np.float64) / 255, (settings["peBlurKernel"],settings["peBlurKernel"]))

        # -------------
        # Track circles
        # -------------
        start = time.perf_counter()
        centers, radii = circFunc(settings["circleTrackingGain"]*(image[:,:,settings["circleTrackingChannel"]] + trackCorrection) * maskArr[:,:,0], settings["guessRadius"], **circleTrackingKwargs)

        # We do some indexing using the centers/radii, so it is helpful
        # to have them as an integer type
        #centers = centers.astype(np.int64)
        #radii = radii.astype(np.int64)

        # We want to keep the order of particles constant, so we make sure
        # that they are (to whatever extent possible) in the same order
        # as the previous frame. This involves finding the closest neighbor
        # from the previous frame.
        if len(centersArr) > 0:
            centerOrder = preserveOrderArgsort(centersArr[-1], centers, padMissingValues=False)
            centers = centers[centerOrder]
            radii = radii[centerOrder]

        
        trackingTimes[i] = time.perf_counter() - start

        # ----------------------
        # Generate initial guess
        # ----------------------
        # We run the initial guess regardless of whether we are going to overwrite
        # with values from the previous frame. This is because the beta values
        # are caluclated via the contact network, which should not be carried over
        # (since the particles are moving).
        forceGuessArr, alphaGuessArr, betaGuessArr = initialForceSolve(peImage,
                                                    centers, radii, settings["fSigma"], settings["pxPerMeter"],
                                                    settings["contactPadding"], settings["g2MaskPadding"],
                                                    contactMaskRadius=settings["contactMaskRadius"],
                                                    boundaryMask=maskArr, ignoreBoundary=ignoreBoundary, g2Cal=g2Cal)

        if len(centersArr) > 0:
            # If we have added/lost particles, we want to carry over the previous values where
            # possible, and otherwise take the results of initialForceSolve

            # Note that this is the complement to the center order calculated previously:
            # this orders the old centers according the new ones.
            # We make the assumption that a particle cannot travel more than it's radius in a single frame
            oldCenterOrder = preserveOrderArgsort(centers, centersArr[-1], padMissingValues=True, maxDistance=settings["guessRadius"])

            # Now find each new particle's old counterpart (if it exists), and then
            # line up the forces using the value of beta, such that we can (optionally)
            # carry over force magnitudes and alpha values.
            for j in range(len(betaGuessArr)):
                if oldCenterOrder[j] is None:
                    continue
                    
                # maxBetaDisplacement should be an angle value (in radians) that a force would
                # never move in a single frame, but is large enough to not lose a force if it
                # moves because of noise/small fluctuations.
                forceOrder = preserveOrderArgsort(betaGuessArr[j], betaArr[-1][oldCenterOrder[j]], padMissingValues=True, maxDistance=settings["maxBetaDisplacement"])

                #print(f'frame {i}, particle {j}: {forceOrder}')

                for k in range(len(forceGuessArr[j])):
                    if forceOrder[k] is not None:
                        if settings["carryOverForce"]:
                            forceGuessArr[j][k] = forceArr[-1][oldCenterOrder[j]][forceOrder[k]]
                        if settings["carryOverAlpha"]:
                            alphaGuessArr[j][k] = alphaArr[-1][oldCenterOrder[j]][forceOrder[k]]


            # In this case, we want to add a small randomly generated contribution
            # so that the algorithm doesn't get stuck in some incorrect loop and so that it
            # explores a little more of the parameter space to find a nice minimum at each step
            if settings["forceNoiseWidth"] is not None:
                forceGuessArr = [np.abs(np.array(f) + np.random.normal(0, settings["forceNoiseWidth"], size=len(f))) for f in forceGuessArr]
            if settings["alphaNoiseWidth"] is not None:
                alphaGuessArr = [np.abs(np.array(a) + np.random.normal(0, settings["alphaNoiseWidth"], size=len(a))) for a in alphaGuessArr]


        initialGuessTimes[i] = time.perf_counter() - trackingTimes[i] - start

        # -------------------------------
        # Optimize each particle's forces
        # -------------------------------
        optimizedForceArr = []
        optimizedBetaArr = []
        optimizedAlphaArr = []
        failed = [False for i in range(len(centers))]

        # Drop forces on any particles whose g2 is lower than the min value
        skipParticles = [False for i in range(len(centers))]
        if checkMinG2:
            gSqr = gSquared(peImage)
            for j in range(len(centers)):
                cMask = circularMask(peImage.shape, centers[j], radii[j])[:,:,0]
                avgG2 = np.sum(gSqr * cMask) / np.sum(cMask)
                skipParticles[j] = avgG2 < minParticleG2

        # Mostly just a debug option, so we can test particle tracking
        if not settings["performOptimization"]:
            optimizedForceArr = forceGuessArr
            optimizedAlphaArr = alphaGuessArr
            optimizedBetaArr = betaGuessArr

        else:
            # This is what should run the majority of the time
            for j in range(len(centers)):
                if not skipParticles[j]:
                    try:
                        # We don't need to pass fSigma, pxPerMeter, or brightfield to the method
                        # because they will get added to optimizationKwargs automatically.
                        optForceArr, optBetaArr, optAlphaArr, res = forceOptimize(forceGuessArr[j], betaGuessArr[j], alphaGuessArr[j], radii[j], centers[j], peImage,
                                                                                  #settings["fSigma"], settings["pxPerMeter"], settings["brightfield"],
                                                                                  **optimizationKwargs)
                        optimizedForceArr.append(optForceArr)
                        optimizedBetaArr.append(optBetaArr)
                        optimizedAlphaArr.append(optAlphaArr)
                    except Exception as ex:
                        print(ex)
                        errorMsgs.append(f'File {imageFiles[i]}: ' + str(ex) + '\n')
                        failed[j] = True
                        totalFailedParticles += 1
                        # Append empty lists (ie say there are no forces) 
                        #optimizedForceArr.append(forceGuessArr[j])
                        #optimizedBetaArr.append(betaGuessArr[j])
                        #optimizedAlphaArr.append(alphaGuessArr[j])
                        optimizedForceArr.append([])
                        optimizedBetaArr.append([])
                        optimizedAlphaArr.append([])
                else:
                    optimizedForceArr.append([])
                    optimizedBetaArr.append([])
                    optimizedAlphaArr.append([])

        # If necessary, impose force balance on all particles
        if requireForceBalance:
            for j in range(len(centers)):
                optimizedForceArr[j], optimizedAlphaArr[j] = singleParticleForceBalance(optimizedForceArr[j], optimizedAlphaArr[j], optimizedBetaArr[j])

        optimizationTimes[i] = time.perf_counter() - initialGuessTimes[i] - trackingTimes[i] - start

        # Save all of our values
        forceArr.append(optimizedForceArr)
        betaArr.append(optimizedBetaArr)
        alphaArr.append(optimizedAlphaArr)
        centersArr.append(centers)
        radiiArr.append(radii)

        if settings["debug"] or settings["saveMovie"] or settings["genFitReport"]:
            estimatedPhotoelasticChannel = np.zeros_like(peImage, dtype=np.float64)    
            for j in range(len(centers)):
                estimatedPhotoelasticChannel += genSyntheticResponse(np.array(forceGuessArr[j]),
                                                                     np.array(alphaGuessArr[j]),
                                                                     np.array(betaGuessArr[j]),
                                                                     settings["fSigma"], radii[j],
                                                                     settings["pxPerMeter"], settings["brightfield"], imageSize=peImage.shape,
                                                                     center=centers[j])
            
            

            optimizedPhotoelasticChannel = np.zeros(peImage.shape)
            for j in range(len(centers)):
                optimizedPhotoelasticChannel += genSyntheticResponse(np.array(optimizedForceArr[j]),
                                                                     np.array(optimizedAlphaArr[j]),
                                                                     np.array(optimizedBetaArr[j]),
                                                                     settings["fSigma"], radii[j],
                                                                     settings["pxPerMeter"], settings["brightfield"], imageSize=peImage.shape,
                                                                     center=centers[j])
            

            # Just simple mean-squared error
            errorArr.append(np.sqrt(np.sum((optimizedPhotoelasticChannel - peImage)**2)))

            imgArr = np.zeros((*optimizedPhotoelasticChannel.shape, 3))
            
            img = Image.fromarray(optimizedPhotoelasticChannel*255)
            img = img.convert('RGB')
            drawObj = ImageDraw.Draw(img)
            for j in range(len(centers)):
                leftUpPoint = (centers[j][1]-radii[j], centers[j][0]-radii[j])
                rightDownPoint = (centers[j][1]+radii[j], centers[j][0]+radii[j])
                twoPointList = [leftUpPoint, rightDownPoint]
                color =  '#FF0000' if failed[j] else '#00AAAA'
                drawObj.ellipse(twoPointList, outline=color, fill=None, width=3)

        if settings["debug"]:

            clear_output(wait=True)
            fig, ax = plt.subplots(1, 3, figsize=(12,4))
            
            ax[0].imshow(maskArr * image)
            ax[0].set_title('Tracked Particles')
            for j in range(len(centers)):
                c = plt.Circle(centers[j][::-1], radii[j], label='Detected particles', color='teal', fill=False, linewidth=1)
                ax[0].add_artist(c)
                # Now add contacts
                for k in range(len(betaGuessArr[j])):
                    contactPoint = centers[j] + radii[j] * np.array([np.cos(betaGuessArr[j][k]), np.sin(betaGuessArr[j][k])])
                    cc = plt.Circle(contactPoint[::-1], 12, color='red', fill=False, linewidth=1)
                    ax[1].add_artist(cc)
                    
                # Now plot past center positions
                #for k in range(len(centersArr)):
                #    if len(centersArr[k]) >= j:
                #        cc = plt.Circle(centersArr[k][j][::-1], 5, color=centerColors[j], fill=True)
                #        ax[0].add_artist(cc)
                    

            ax[1].imshow(estimatedPhotoelasticChannel)
            ax[1].set_title('Initial Guess for Optimizer\n(known forces)')
            
            
            ax[2].imshow(img)
            ax[2].set_title('Optimized Forces\n(known forces)')
            
            fig.suptitle(imageFiles[i])
            fig.tight_layout()
            plt.show()

        if settings["saveMovie"]:
            imageArr.append(img)

        miscTimes[i] = time.perf_counter() - optimizationTimes[i] - initialGuessTimes[i] - trackingTimes[i] - start 

        if settings["debug"]: 
            print(f'Took {time.perf_counter() - start:.5}s to solve frame:')
            print(f'{5*" "}Tracking:         {trackingTimes[i]:.3}s')
            print(f'{5*" "}Initial guess:    {initialGuessTimes[i]:.3}s')
            print(f'{5*" "}Optimization:     {optimizationTimes[i]:.3}s')
            print(f'{5*" "}Misc. processes:  {miscTimes[i]:.3}s')

        if settings["showProgressBar"]:
            bar.update()
   
    
    # Restructure the arrays to make them more friendly, and to track forces/particles across timesteps
    rectForceArr, rectAlphaArr, rectBetaArr, rectCenterArr, rectRadiusArr = rectangularizeForceArrays(forceArr, alphaArr, betaArr, centersArr, radiiArr)

    # --------------
    # Track rotation
    # --------------
    # We choose to do this after the actual solving because it helps
    # to have the rectangular force arrays.
    padding = settings["guessRadius"] + 5

    # First, we generate our reference images, which are the first
    # time a particle is completely in frame.
    refImages = [None] * len(rectCenterArr)
    for i in range(len(refImages)):
        for j in range(len(imageFiles)):
            if not True in np.isnan(rectCenterArr[i][j]):
                # Continue to the next frame if this one is partially offscreen
                if True in ((rectCenterArr[i][j] - padding) < 0) or True in ((rectCenterArr[i][j] - np.array(imageSize) + padding) > 0):
                    continue

                # Otherwise, this is a good frame, so we save it
                refImageFull = checkImageType(settings["imageDirectory"] + imageFiles[j])[:,xB[0]:xB[1],settings["circleTrackingChannel"]]
                refImageFull *= circularMask(refImageFull.shape, rectCenterArr[i][j], rectRadiusArr[i][j])[:,:,0]

                refImages[i] = refImageFull[int(rectCenterArr[i][j][0] - padding):int(rectCenterArr[i][j][0] + padding), int(rectCenterArr[i][j][1] - padding):int(rectCenterArr[i][j][1] + padding)]
                # And move onto the next particle
                break

    # Same shape as the radius array: 1 value for each timestep, for each particle
    rectAngleArr = np.zeros(rectRadiusArr.shape)
    # Set all values to be np.nan initially
    rectAngleArr[:,:] = np.nan
    
    # Now we compare that reference particle to each subsequent frame
    # (probably not best practice that I've switched the indices
    # with respect to the previous statements, but :/)
    for i in range(len(imageFiles)):
        currentImageFull = checkImageType(settings["imageDirectory"] + imageFiles[i])[:,xB[0]:xB[1],settings["circleTrackingChannel"]]

        for j in range(len(refImages)):
            # Make sure we have a reference image, and the particle is in full view
            if True in np.isnan(rectCenterArr[j][i]):
                continue

            if True in ((rectCenterArr[j][i] - padding) < 0) or True in ((rectCenterArr[j][i] - np.array(imageSize) + padding) > 0):
                continue

            # Crop out around the particle and mask it
            currImage = (circularMask(currentImageFull.shape, rectCenterArr[j][i], rectRadiusArr[j][i])[:,:,0] * currentImageFull)[int(rectCenterArr[j][i][0] - padding):int(rectCenterArr[j][i][0] + padding), int(rectCenterArr[j][i][1] - padding):int(rectCenterArr[j][i][1] + padding)]
           
            # Which is the kernel and which is the reference image doesn't really matter
            # (as long as we are consistent)
            # We can choose our bounds based on the previous value of the rotation
            if i >= 1 and not np.isnan(rectAngleArr[j,i-1]):
                rotationBounds = (rectAngleArr[j,i-1] - .1, rectAngleArr[j,i-1] + .1)
            else:
                # If either i=0 or the previous rotation value is nan, we should start around 0
                # anyway (since we define 0 arbitrarily)
                rotationBounds = (-.2, .2)

            # .003 was chosen based on the data presented in the wiki
            # https://github.com/Jfeatherstone/pepe/wiki/Angular-Convolution
            thetaArr, convArr = angularConvolution(refImages[j], currImage, dTheta=.003, angleBounds=rotationBounds)
            rectAngleArr[j,i] = thetaArr[findPeaksMulti(convArr)[0][0][0]]


    # Reuse the name of the folder the images come from as a part of
    # the output folder name
    # [-2] element for something of form 'path/to/final/folder/' will be 'folder'
    # If we are missing the final /, you have to take just the [-1] element
    if settings["imageDirectory"][-1] == '/':
        outputFolderPath = outputRootFolder + settings["imageDirectory"].split('/')[-2] + f'_Synthetic{settings["outputExtension"]}/'
    else:
        outputFolderPath = outputRootFolder + settings["imageDirectory"].split('/')[-1] + f'_Synthetic{settings["outputExtension"]}/'

    if not os.path.exists(outputFolderPath):
        os.mkdir(outputFolderPath)

    if settings["saveMovie"]:
        imageArr[0].save(outputFolderPath + 'Synthetic.gif', save_all=True, append_images=imageArr[1:], duration=30, optimize=False, loop=0)

    # Write a readme file that contains all of the parameters that the solving used
    lines = ['#####################\n',
             '#    README FILE    #\n',
             '#####################\n']

    lines += [f'Generated: {time.ctime()}\n\n']
    lines += ['Note: this file was autogenerated by the `pepe.auto.forceSolve()` function\n',
              '      and it is not recommended to be manually edited. To reuse the settings\n',
              '      and parameters that were used here, the path of this file\n',
             f'      (\'{outputFolderPath}readme.txt\') \n',
              '      can be passed via the \'settingsFile\' keyword argument of `pepe.auto.forceSolve()`.\n',
              '      In this case, explictly passed arguments will override the values in the settings file.\n']

    lines += ['\n## Runtime Information\n',
              f'Version: pepe {pepe.__version__}\n',
              f'Total runtime: {time.perf_counter() - overallStartTime:.6}s\n',
              f'Mean tracking time: {np.mean(trackingTimes):.4}s\n',
              f'Mean guess generation time: {np.mean(initialGuessTimes):.4}s\n',
              f'Mean optimization time: {np.mean(optimizationTimes):.4}s\n',
              f'Mean misc. time: {np.mean(miscTimes):.4}s\n',
              f'Number of failed particles: {totalFailedParticles}\n']

    settings.update(circleTrackingKwargs)
    settings.update(optimizationKwargs)

    lines += ['\n## Settings\n']
    for k,v in settings.items():
        lines += [f'{k}: {v}\n']

    lines += ['\n## Errors\n']
    if len(errorMsgs) > 0:
        lines += errorMsgs
    else:
        lines += ['None :)']

    with open(outputFolderPath + 'readme.txt', 'w') as readmeFile:
        readmeFile.writelines(lines)

    # Save the arrays to pickle files (optional)
    if settings["pickleArrays"]:
        with open(outputFolderPath + 'forces.pickle', 'wb') as f:
            pickle.dump(rectForceArr, f)

        with open(outputFolderPath + 'alphas.pickle', 'wb') as f:
            pickle.dump(rectAlphaArr, f)

        with open(outputFolderPath + 'betas.pickle', 'wb') as f:
            pickle.dump(rectBetaArr, f)

        with open(outputFolderPath + 'centers.pickle', 'wb') as f:
            pickle.dump(rectCenterArr, f)

        with open(outputFolderPath + 'radii.pickle', 'wb') as f:
            pickle.dump(rectRadiusArr, f)

        with open(outputFolderPath + 'angles.pickle', 'wb') as f:
            pickle.dump(rectAngleArr, f)

    # Save the raw arrays too, since I think I have a bug in my rectangularization process
#    if settings["pickleArrays"]:
#        with open(outputFolderPath + 'forces_raw.pickle', 'wb') as f:
#            pickle.dump(forceArr, f)
#
#        with open(outputFolderPath + 'alphas_raw.pickle', 'wb') as f:
#            pickle.dump(alphaArr, f)
#
#        with open(outputFolderPath + 'betas_raw.pickle', 'wb') as f:
#            pickle.dump(betaArr, f)
#
#        with open(outputFolderPath + 'centers_raw.pickle', 'wb') as f:
#            pickle.dump(centersArr, f)
#
#        with open(outputFolderPath + 'radii_raw.pickle', 'wb') as f:
#            pickle.dump(radiiArr, f)

    # Generate a fit report (optional)
    # This include informtaion about the error for each frame, all of the forces/alphas/betas/
    # centers/radii for each particle at each timestep, and all settings in a nicely compiled
    # (via latex) pdf.
    if settings["genFitReport"]:
        # Make the source directory
        if not os.path.exists(outputFolderPath + 'FitReport_src'):
            os.mkdir(outputFolderPath + 'FitReport_src')

        # First, generate a plot of the error
        fig, ax = plt.subplots()

        ax.plot(errorArr)
        ax.set_xlabel('Frame')
        ax.set_ylabel('Mean-squared error')
        ax.set_title('Difference Between Optimized Result and Real Image')

        fig.savefig(outputFolderPath + 'FitReport_src/error.pdf')
        fig.savefig(outputFolderPath + 'FitReport_src/error.png')
        plt.close(fig)

        # Draw all of the circles, with their labeled numbers
        fig, ax = plt.subplots(1, 2, figsize=(8,3))
        # First timestep
        visCircles([rectCenterArr[i][0] for i in range(len(rectCenterArr))], [rectRadiusArr[i][0] for i in range(len(rectRadiusArr))],
                   ax=ax[0], annotations=np.arange(len(rectCenterArr)), setBounds=True) 
        # Last timestep
        visCircles([rectCenterArr[i][-1] for i in range(len(rectCenterArr))], [rectRadiusArr[i][-1] for i in range(len(rectRadiusArr))],
                   ax=ax[1], annotations=np.arange(len(rectCenterArr)), setBounds=True) 

        for i in range(2):
            ax[i].set_xlabel('X [px]')
            ax[i].set_ylabel('Y [px]')
            ax[i].invert_yaxis()

        ax[0].set_title('First Frame')
        ax[1].set_title('Last Frame')

        fig.savefig(outputFolderPath + 'FitReport_src/particle_identities.pdf')
        fig.savefig(outputFolderPath + 'FitReport_src/particle_identities.png')
        plt.close(fig)

        # Next, draw the forces/betas/alphas/centers for each particle
        # through time
        for i in range(len(rectForceArr)):
            fig, ax = visForces(rectForceArr[i], rectAlphaArr[i], rectBetaArr[i], rectCenterArr[i], rectAngleArr[i])
            fig.suptitle(f'Particle {i}')
            fig.savefig(outputFolderPath + f'FitReport_src/particle_{i}_forces.pdf')
            fig.savefig(outputFolderPath + f'FitReport_src/particle_{i}_forces.png')
            plt.close(fig)

        # Create a gif of the particle orientation through time, overlaid
        # on the original images
        visRotation([settings["imageDirectory"] + f for f in imageFiles],
                   rectCenterArr, rectRadiusArr, rectAngleArr, outputFolderPath + 'FitReport_src/', (0, cropXMin))

        # Create gifs of the contacts
        forceColors = genColors(len(rectBetaArr))
        # The list comprehension is to make sure that we index a particle that actually has forces acting
        # on it.
        tSteps = len(imageFiles)#len([b for b in rectBetaArr if len(b) > 0][0])
        contactPointImages = [None for i in range(tSteps)]
        contactAngleImages = [None for i in range(tSteps)]

        for i in range(tSteps):
            # Have to do this, because the settings variable could be None 
            startI = settings["imageStartIndex"] if settings["imageStartIndex"] is not None else 0

            # First, just the contact points
            fig, ax = plt.subplots()
            visCircles([rectCenterArr[p][i] for p in range(len(rectCenterArr))], [rectRadiusArr[p][i] for p in range(len(rectRadiusArr))], ax=ax)

            for particleIndex in range(len(rectBetaArr)):
                visContacts(rectCenterArr[particleIndex][i], rectRadiusArr[particleIndex][i],
                            rectBetaArr[particleIndex][:,i], ax=ax, forceColors=forceColors[particleIndex])

            ax.set_xlim([0, 1280])
            ax.set_ylim([0, 1024])
            ax.set_aspect('equal')
            ax.set_title(f'Frame {i + startI}')
            ax.invert_yaxis()

            canvas = plt.get_current_fig_manager().canvas
            canvas.draw()
            contactPointImages[i] = Image.frombytes('RGB', canvas.get_width_height(), 
                                        canvas.tostring_rgb())

            plt.close(fig)
       
            # Now the one with angles
            fig, ax = plt.subplots()
            visCircles([rectCenterArr[p][i] for p in range(len(rectCenterArr))], [rectRadiusArr[p][i] for p in range(len(rectRadiusArr))], ax=ax)

            for particleIndex in range(len(rectBetaArr)):
                visContacts(rectCenterArr[particleIndex][i], rectRadiusArr[particleIndex][i],
                            rectBetaArr[particleIndex][:,i], ax=ax, forceColors=forceColors[particleIndex], alphaArr=rectAlphaArr[particleIndex][:,i])
                                                
            ax.set_xlim([0, 1280])
            ax.set_ylim([0, 1024])
            ax.set_aspect('equal')
            ax.set_title(f'Frame {i + startI}')
            ax.invert_yaxis()

            canvas = plt.get_current_fig_manager().canvas
            canvas.draw()
            contactAngleImages[i] = Image.frombytes('RGB', canvas.get_width_height(), 
                                        canvas.tostring_rgb())

            plt.close(fig)

        contactPointImages[0].save(outputFolderPath + 'FitReport_src/contact_points.gif', save_all=True,
                                   append_images=contactPointImages[1:], duration=20, optimize=True, loop=0)

        contactAngleImages[0].save(outputFolderPath + 'FitReport_src/contact_angles.gif', save_all=True,
                                   append_images=contactAngleImages[1:], duration=20, optimize=True, loop=0)

    if settings["showProgressBar"]:
        bar.update()
        bar.close()

    return rectForceArr, rectAlphaArr, rectBetaArr, rectCenterArr, rectRadiusArr, rectAngleArr



import numpy as np
from scipy.stats import linregress

def determineLinearRegions(data, minLength=.1, minR2=.96, maxSlopeInterceptDiff=.75):
    """
    Determine regions of a plot that are approximately linear by performing
    linear least-squares on a rolling window.
    
    Parameters
    ----------
    
    data : array_like
        Data within which linear regions are to be identified
        
    minLength : int or float
        The minimum length of a linear segment, either as an
        integer number of indices, or as a float fraction of the
        overall data length.
        
    minR2 : float
        The minimum r-squared value for a region to be
        considered linear.
        
    maxSlopeInterceptDiff : float
        The float percentage difference allowed between slopes
        and intercepts of adjacent slices for them to be
        considered the same region.
        
    Returns
    -------
    
    regionIndices : np.ndarray[N,2]
        The start and end indices for the N detected regions.
        
    slopes : np.ndarray[N]
        The slope of each region.
        
    intercepts : np.ndarray[N]
        The intercept of each region.
    """
    if minLength < 1:
        minLinSteps = int(len(data)*minLength)
    else:
        minLinSteps = int(minLength)
        
    inLinearRegion = False
    linearRegions = []
    slopes = []
    intercepts = []
    
    # Perform least squares on a rolling window
    i = 0
    while i < len(data) - minLinSteps:
        xArr = np.arange(i, i+minLinSteps)
        slope, intercept, r2, p_value, std_err = linregress(xArr, data[i:i+minLinSteps])
        
        if np.abs(r2) > minR2:
            
            if inLinearRegion:
                # Calculate how different new slope is from old one
                if np.abs((np.mean(slopes[-1]) - slope) / np.mean(slopes[-1])) < maxSlopeInterceptDiff and np.abs((np.mean(intercepts[-1]) - intercept) / np.mean(intercepts[-1])) < maxSlopeInterceptDiff:
                    # This is still the same linear region, so we extend the bounds
                    linearRegions[-1][1] = i+minLinSteps
                    
                    # And average in the slopes and intercepts
                    slopes[-1] += [slope]
                    intercepts[-1] += [intercept]
                else:
                    # Otherwise, we have a new linear region, which we start
                    # at the end of the other one
                    i = linearRegions[-1][1]
                    inLinearRegion = False
                    continue

            else:
                # New linear region
                linearRegions.append([i, i+minLinSteps])
                slopes.append([slope])
                intercepts.append([intercept])
                
            inLinearRegion = True
        
        else:
            inLinearRegion = False
            
        i += 1
            
    slopes = np.array([np.mean(s) for s in slopes])
    intercepts = np.array([np.mean(inter) for inter in intercepts])
    return np.array(linearRegions), slopes, intercepts
